// vincia04-root is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to study n-jet rates and the so-called
// 4-jet angles LEP 1, using VINCIA's ROOT-based runtime display. 
// Uses VINCIA for showering and matching, FASTJET for jet clustering, 
// and PYTHIA 8 for all else
// Copyright (C) 2019 Peter Skands

// PYTHIA and VINCIA include files
#include "Pythia8/Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "LiveDisplays/LiveDisplays.h"
#include "TFile.h"
// FASTJET
#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 

// Auxiliary routine to compute angles between jets
double jetAngle(fastjet::PseudoJet jet1, fastjet::PseudoJet jet2) {
    Vec4 p1(jet1.px(),jet1.py(),jet1.pz(),jet1.e());
    Vec4 p2(jet2.px(),jet2.py(),jet2.pz(),jet2.e());
    return theta(p1,p2);
}

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects
  Pythia pythia; 
  VinciaPlugin vincia(&pythia);

  // Read user settings from file
  pythia.readFile("vincia04.cmnd");

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Shorthand for event record
  Event& event = pythia.event;

  // Some useful pointers
  Settings* settingsPtr = vincia.settingsPtr;        // to settings database

  // Extract settings to be used in the main program.  
  double eCM       = settingsPtr->parm("Beams:eCM");
  int    nEvent    = settingsPtr->mode("Main:numberOfEvents");

  // Vincia settings
  bool   vinciaOn         = settingsPtr->flag("Vincia");
  int    verbose          = settingsPtr->mode("Vincia:verbose");

  // Example of extracting a mass value:
  // double bMass            = vincia.particleDataPtr->m0(5);

  // hadronization settings
  bool   hadronLevel      = settingsPtr->flag("HadronLevel:all");

  // Data path directory, extract from VINCIADATA
  string VINCIA_SHARE     = vincia.VINCIA_SHARE;
  string VINCIA_VAL       = VINCIA_SHARE+"validation/";
  string VINCIA_DAT       = VINCIA_SHARE+"data/";

  //************************************************************************
  // LiveDisplays
  //************************************************************************

  if (!vinciaOn) {
    settingsPtr->word("LiveDisplays:legendLabel","Pythia");
  } else {
    settingsPtr->word("LiveDisplays:legendLabel","Vincia 2.3");
  }

  // Create a LiveApp: initialises & handles graphics communications
  LiveDisplaysApp liveDisplaysApp(&pythia);

  // Read files
  vector<string> mcFileRead;
  mcFileRead.push_back(string(VINCIA_VAL+"vincia2/vincia04").c_str());
  mcFileRead.push_back(string(VINCIA_VAL+"pythia/vincia04").c_str());
 
  // Prefix of rescue files (to initialise from last run)
  string rescuePrefix = "vincia05-";
  
  // Next event at which LiveDisplaysApp will save displays/histograms
  int iSave = liveDisplaysApp.nNextSave();

  //************************************************************************
  // Initialize VINCIA (automatically initializes PYTHIA) 
  //************************************************************************

  // Initialize
  if(!vincia.init()) { return EXIT_FAILURE; }

  //***************************************************************************
  // FASTJET
  //***************************************************************************

  // Fastjet analysis - select algorithm and parameters
  fastjet::Strategy               strategy = fastjet::Best;
  fastjet::RecombinationScheme    recombScheme = fastjet::E_scheme;
  fastjet::JetDefinition         *jetDef = NULL;
  jetDef = new fastjet::JetDefinition(fastjet::ee_kt_algorithm,
                                      recombScheme, strategy);
  // Fastjet input
  std::vector <fastjet::PseudoJet> fjInputs;
 
  //***************************************************************************
  // ROOT
  //***************************************************************************

  // Form name for ROOT file
  ostringstream outfile;
  // Put root files in subdirectory defined at beginning of file
  outfile<<"vincia04.root";

  //****************
  // Book Histograms

  // Vincia uncertainty bands: number of variations expected 
  int nWeights = max(1,vincia.nWeights());

  // Jet Rates
  int nBins  = 6;
  double xLo = -1.2;
  double xHi =  0.0;
  vector<TH1D*> Vr1j;
  Vr1j.push_back(new TH1D("r1j","Exclusive 1-Jet Rate",nBins, xLo, xHi));
  // Add to display with
  // logy = true, normalize=false, showStatErr=true
  Vr1j[0]->SetMaximum(10.); 
  Vr1j[0]->SetMinimum(0.00001); 
  Vr1j[0]->GetXaxis()->SetTitle("Ln(y1)");
  Vr1j[0]->GetYaxis()->SetTitle("Rate");
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vr1j.push_back(new TH1D(string("r1j_"+num2str(iWeight,1)).c_str()," ",
                            nBins,xLo,xHi));  
  LiveDisplay dispR1j(Vr1j,&pythia);
  dispR1j.setDataFile(string(VINCIA_DAT+"Aleph-91-r1j.dat").c_str(),
    10,30,40,51,52,61,62,71,72);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR1j.addHisFile(mcFileRead[iFile]+"-r1j.dat");
  dispR1j.normalize(false);
  dispR1j.logY(true);

  // Jet Rates
  nBins = 40;
  xLo   =  -8.0;
  xHi   =   0.0;
  vector<TH1D*> Vr2j;
  Vr2j.push_back(new TH1D("r2j","Hard Exclusive 2-Jet Rate",nBins,xLo,xHi));
  Vr2j[0]->SetMaximum(50.01); 
  Vr2j[0]->SetMinimum(0.005); 
  Vr2j[0]->GetXaxis()->SetTitle("Ln(y2)");
  Vr2j[0]->GetYaxis()->SetTitle("Rate");
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vr2j.push_back(new TH1D(string("r2j_"+num2str(iWeight,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  LiveDisplay dispR2j(Vr2j,&pythia);
  dispR2j.setDataFile(string(VINCIA_DAT+"Aleph-91-r2j.dat").c_str(),
    10,30,40,51,52,61,62,71,72);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR2j.addHisFile(mcFileRead[iFile]+"-r2j.dat");
  dispR2j.normalize(false); 
  dispR2j.logY(true);

  // Jet Rates
  nBins = 11;
  xLo   = -10.2;
  xHi   =  -8.0;
  vector<TH1D*> Vr2jSoft;
  Vr2jSoft.push_back(new TH1D("r2jSoft","Soft Exclusive 2-Jet Rate",nBins,xLo,xHi));
  Vr2jSoft[0]->SetMaximum(1.); 
  Vr2jSoft[0]->SetMinimum(0.000001); 
  Vr2jSoft[0]->GetXaxis()->SetTitle("Ln(y2)");
  Vr2jSoft[0]->GetYaxis()->SetTitle("Rate");
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vr2jSoft.push_back(new TH1D(string("r2jSoft_"+num2str(iWeight,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  LiveDisplay dispR2jSoft(Vr2jSoft,&pythia);
  dispR2jSoft.setDataFile(string(VINCIA_DAT+"Aleph-91-r2j.dat").c_str(),
    10,30,40,51,52,61,62,71,72);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR2jSoft.addHisFile(mcFileRead[iFile]+"-r2jSoft.dat");
  dispR2jSoft.normalize(false); 
  dispR2jSoft.logY(true);

  // Jet Rates
  nBins =    35;
  xLo   =  -8.0;
  xHi   =  -1.0;
  vector<TH1D*> Vr3j;
  Vr3j.push_back(new TH1D("r3j","Hard Exclusive 3-Jet Rate",nBins,xLo,xHi));
  Vr3j[0]->SetMaximum(50.01); 
  Vr3j[0]->SetMinimum(0.0005); 
  Vr3j[0]->GetXaxis()->SetTitle("Ln(y3)");
  Vr3j[0]->GetYaxis()->SetTitle("Rate");
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vr3j.push_back(new TH1D(string("r3j_"+num2str(iWeight,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  LiveDisplay dispR3j(Vr3j,&pythia);
  dispR3j.setDataFile(string(VINCIA_DAT+"Aleph-91-r3j.dat").c_str(),
    10,30,40,51,52,61,62,71,72);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR3j.addHisFile(mcFileRead[iFile]+"-r3j.dat");
  dispR3j.normalize(false); 
  dispR3j.logY(true);

  // Jet Rates
  nBins =    11;
  xLo   = -10.2;
  xHi   =  -8.0;
  vector<TH1D*> Vr3jSoft;
  Vr3jSoft.push_back(new TH1D("r3jSoft","Soft Exclusive 3-Jet Rate",
      nBins,xLo,xHi));
  Vr3jSoft[0]->SetMaximum(2.01); 
  Vr3jSoft[0]->SetMinimum(0.00001); 
  Vr3jSoft[0]->GetXaxis()->SetTitle("Ln(y3)");
  Vr3jSoft[0]->GetYaxis()->SetTitle("Rate");
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vr3jSoft.push_back(new TH1D(string("r3jSoft_"+num2str(iWeight,1)).c_str(),
        " ",nBins,xLo,xHi));  
  LiveDisplay dispR3jSoft(Vr3jSoft,&pythia);
  dispR3jSoft.setDataFile(string(VINCIA_DAT+"Aleph-91-r3j.dat").c_str(),
    10,30,40,51,52,61,62,71,72);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR3jSoft.addHisFile(mcFileRead[iFile]+"-r3jSoft.dat");
  dispR3jSoft.normalize(false); 
  dispR3jSoft.logY(true);

  // Jet Rates
  nBins =    30;
  xLo   =  -8.0;
  xHi   =  -2.0;
  vector<TH1D*> Vr4j;
  Vr4j.push_back(new TH1D("r4j","Hard Exclusive 4-Jet Rate",nBins,xLo,xHi));
  Vr4j[0]->SetMaximum(50.); 
  Vr4j[0]->SetMinimum(0.000001); 
  Vr4j[0]->GetXaxis()->SetTitle("Ln(y4)");
  Vr4j[0]->GetYaxis()->SetTitle("Rate");
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vr4j.push_back(new TH1D(string("r4j_"+num2str(iWeight,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  LiveDisplay dispR4j(Vr4j,&pythia);
  dispR4j.setDataFile(string(VINCIA_DAT+"Aleph-91-r4j.dat").c_str(),
    10,30,40,51,52,61,62,71,72);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR4j.addHisFile(mcFileRead[iFile]+"-r4j.dat");
  dispR4j.normalize(false); 
  dispR4j.logY(true);

  // Jet Rates
  nBins =    11;
  xLo   = -10.2;
  xHi   =  -8.0;
  vector<TH1D*> Vr4jSoft;
  Vr4jSoft.push_back(new TH1D("r4jSoft","Soft Exclusive 4-Jet Rate",
      nBins,xLo,xHi));
  Vr4jSoft[0]->SetMaximum(2.); 
  Vr4jSoft[0]->SetMinimum(0.0001); 
  Vr4jSoft[0]->GetXaxis()->SetTitle("Ln(y4)");
  Vr4jSoft[0]->GetYaxis()->SetTitle("Rate");
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vr4jSoft.push_back(new TH1D(string("r4jSoft_"+num2str(iWeight,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  LiveDisplay dispR4jSoft(Vr4jSoft,&pythia);
  dispR4jSoft.setDataFile(string(VINCIA_DAT+"Aleph-91-r4j.dat").c_str(),
    10,30,40,51,52,61,62,71,72);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR4jSoft.addHisFile(mcFileRead[iFile]+"-r4jSoft.dat");
  dispR4jSoft.normalize(false); 
  dispR4jSoft.logY(true);

  // Jet Rates
  nBins =    25;
  xLo   =  -8.0;
  xHi   =  -3.0;
  vector<TH1D*> Vr5j;
  Vr5j.push_back(new TH1D("r5j","Hard Exclusive 5-Jet Rate",nBins,xLo,xHi));
  Vr5j[0]->SetMaximum(50.); 
  Vr5j[0]->SetMinimum(0.000001); 
  Vr5j[0]->GetXaxis()->SetTitle("Ln(y5)");
  Vr5j[0]->GetYaxis()->SetTitle("Rate");
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vr5j.push_back(new TH1D(string("r5j_"+num2str(iWeight,1)).c_str()," ",
                                    nBins,xLo,xHi));  
LiveDisplay dispR5j(Vr5j,&pythia);
  dispR5j.setDataFile(string(VINCIA_DAT+"Aleph-91-r5j.dat").c_str(),
    10,30,40,51,52,61,62,71,72);
for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR5j.addHisFile(mcFileRead[iFile]+"-r5j.dat");
  dispR5j.normalize(false); 
  dispR5j.logY(true);

  // Jet Rates
  nBins =    11;
  xLo   = -10.2;
  xHi   =  -8.0;
  vector<TH1D*> Vr5jSoft;
  Vr5jSoft.push_back(new TH1D("r5jSoft","Soft Exclusive 5-Jet Rate",nBins,xLo,xHi));
  Vr5jSoft[0]->SetMaximum(2.0); 
  Vr5jSoft[0]->SetMinimum(0.001); 
  Vr5jSoft[0]->GetXaxis()->SetTitle("Ln(y5)");
  Vr5jSoft[0]->GetYaxis()->SetTitle("Rate");
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vr5jSoft.push_back(new TH1D(string("r5jSoft_"+num2str(iWeight,1)).c_str(),
        " ",nBins,xLo,xHi));  
  LiveDisplay dispR5jSoft(Vr5jSoft,&pythia);
  dispR5jSoft.setDataFile(string(VINCIA_DAT+"Aleph-91-r5j.dat").c_str(),
    10,30,40,51,52,61,62,71,72);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR5jSoft.addHisFile(mcFileRead[iFile]+"-r5jSoft.dat");
  dispR5jSoft.normalize(false); 
  dispR5jSoft.logY(true);

  // Jet Rates
  nBins =    31;
  xLo   = -10.2;
  xHi   =  -4.0;
  vector<TH1D*> Vr6j;
  Vr6j.push_back(new TH1D("r6j","Inclusive 6-Jet Rate",nBins,xLo,xHi));
  Vr6j[0]->SetMaximum(50.01); 
  Vr6j[0]->SetMinimum(0.000005); 
  Vr6j[0]->GetXaxis()->SetTitle("Ln(y6)");
  Vr6j[0]->GetYaxis()->SetTitle("Rate");
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vr6j.push_back(new TH1D(string("r6j_"+num2str(iWeight,1)).c_str()," ",
                                    nBins,xLo,xHi));  
  LiveDisplay dispR6j(Vr6j,&pythia);
  dispR6j.setDataFile(string(VINCIA_DAT+"Aleph-91-r6j.dat").c_str(),
    10,30,40,51,52,61,62,71,72);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR6j.addHisFile(mcFileRead[iFile]+"-r6j.dat");
  dispR6j.normalize(false); 
  dispR6j.logY(true);

  // Jet Resolutions
  nBins = 13;
  xLo   = 0.0;
  xHi   = 1.3;
  double wY12 = nBins/(xHi-xLo);
  vector<TH1D*> VLnY12;
  VLnY12.push_back(new TH1D("LnY12","Ln(1/y12)",nBins, xLo, xHi));
  VLnY12[0]->SetMaximum(100.0); 
  VLnY12[0]->SetMinimum(0.000001); 
  VLnY12[0]->GetXaxis()->SetTitle("Ln(1/y12)");
  VLnY12[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++) 
     VLnY12.push_back(new TH1D(string("LnY12_"+num2str(iWeight,1)).c_str()," ",
                                          nBins,xLo,xHi));  
  LiveDisplay dispLnY12(VLnY12,&pythia);
  dispLnY12.setDataFile(string(VINCIA_DAT+"Aleph-91-LnY12.dat").c_str(),
    10,30,40,50,60,70);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispLnY12.addHisFile(mcFileRead[iFile]+"-LnY12.dat");
  dispLnY12.normalize(true);
  dispLnY12.logY(true);
  
  // Jet Resolutions
  nBins = 36;
  xLo = 0.8;
  xHi = 8.0;
  double wY23 = nBins/(xHi-xLo);
  vector<TH1D*> VLnY23;
  VLnY23.push_back(new TH1D("LnY23","Ln(1/y23) (Hard)",nBins, xLo, xHi));
  VLnY23[0]->SetMaximum(50.); 
  VLnY23[0]->SetMinimum(0.0005); 
  VLnY23[0]->GetXaxis()->SetTitle("Ln(1/y23)");
  VLnY23[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VLnY23.push_back(new TH1D(string("LnY23_"+num2str(iWeight,1)).c_str()," ",
                                          nBins,xLo,xHi));  
  LiveDisplay dispLnY23(VLnY23,&pythia);
  dispLnY23.setDataFile(string(VINCIA_DAT+"Aleph-91-LnY23.dat").c_str(),
    10,30,40,50,60,70);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispLnY23.addHisFile(mcFileRead[iFile]+"-LnY23.dat");
  dispLnY23.normalize(true);
  dispLnY23.logY(true);

  nBins = 15;
  xLo = 8.0;
  xHi = 11.0;
  double wY23soft = nBins/(xHi-xLo);
  vector<TH1D*> VLnY23soft;
  VLnY23soft.push_back(new TH1D("LnY23soft","Ln(1/y23) (Soft)",nBins,xLo,xHi));
  VLnY23soft[0]->SetMaximum(10.); 
  VLnY23soft[0]->SetMinimum(0.00001); 
  VLnY23soft[0]->GetXaxis()->SetTitle("Ln(1/y23)");
  VLnY23soft[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VLnY23soft.push_back(new TH1D(string("LnY23soft_"+num2str(iWeight,1)).c_str()," ",nBins,xLo,xHi));  
  LiveDisplay dispLnY23soft(VLnY23soft,&pythia);
  dispLnY23soft.setDataFile(string(VINCIA_DAT+"Aleph-91-LnY23.dat").c_str(),
    10,30,40,50,60,70);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispLnY23soft.addHisFile(mcFileRead[iFile]+"-LnY23soft.dat");
  dispLnY23soft.normalize(true);
  dispLnY23soft.logY(true);

  nBins = 31;
  xLo = 1.8;
  xHi = 8.0;
  double wY34 = nBins/(xHi-xLo);
  vector<TH1D*> VLnY34;
  VLnY34.push_back(new TH1D("LnY34","Ln(1/y34) (Hard)",nBins, xLo, xHi));
  VLnY34[0]->SetMaximum(50.); 
  VLnY34[0]->SetMinimum(0.00001); 
  VLnY34[0]->GetXaxis()->SetTitle("Ln(1/y34)");
  VLnY34[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VLnY34.push_back(new TH1D(string("LnY34_"+num2str(iWeight,1)).c_str()," ",
                                          nBins,xLo,xHi)); 
  LiveDisplay dispLnY34(VLnY34,&pythia);
  dispLnY34.setDataFile(string(VINCIA_DAT+"Aleph-91-LnY34.dat").c_str(),
    10,30,40,50,60,70);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispLnY34.addHisFile(mcFileRead[iFile]+"-LnY34.dat");
  dispLnY34.normalize(true);
  dispLnY34.logY(true);

  nBins = 15;
  xLo = 8.0;
  xHi = 11.0;
  double wY34soft = nBins/(xHi-xLo);
  vector<TH1D*> VLnY34soft;
  VLnY34soft.push_back(new TH1D("LnY34soft","Ln(1/y34) (Soft)",nBins,xLo,xHi));
  VLnY34soft[0]->SetMaximum(10.); 
  VLnY34soft[0]->SetMinimum(0.00001); 
  VLnY34soft[0]->GetXaxis()->SetTitle("Ln(1/y34)");
  VLnY34soft[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VLnY34soft.push_back(new TH1D(string("LnY34soft_"+num2str(iWeight,1)).c_str()," ",nBins,xLo,xHi)); 
  LiveDisplay dispLnY34soft(VLnY34soft,&pythia);
  dispLnY34soft.setDataFile(string(VINCIA_DAT+"Aleph-91-LnY34.dat").c_str(),
    10,30,40,50,60,70);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispLnY34soft.addHisFile(mcFileRead[iFile]+"-LnY34soft.dat");
  dispLnY34soft.normalize(true);
  dispLnY34soft.logY(true);

  nBins = 25;
  xLo = 3.0;
  xHi = 8.0;
  double wY45 = nBins/(xHi-xLo);
  vector<TH1D*> VLnY45;
  VLnY45.push_back(new TH1D("LnY45","Ln(1/y45) (Hard)",nBins, xLo, xHi));
  VLnY45[0]->SetMaximum(50.); 
  VLnY45[0]->SetMinimum(0.00001); 
  VLnY45[0]->GetXaxis()->SetTitle("Ln(1/y45)");
  VLnY45[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VLnY45.push_back(new TH1D(string("LnY45_"+num2str(iWeight,1)).c_str()," ",
                                          nBins,xLo,xHi));  
  LiveDisplay dispLnY45(VLnY45,&pythia);
  dispLnY45.setDataFile(string(VINCIA_DAT+"Aleph-91-LnY45.dat").c_str(),
    10,30,40,50,60,70);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispLnY45.addHisFile(mcFileRead[iFile]+"-LnY45.dat");
  dispLnY45.normalize(true);
  dispLnY45.logY(true);

  nBins = 15;
  xLo = 8.0;
  xHi = 11.0;
  double wY45soft = nBins/(xHi-xLo);
  vector<TH1D*> VLnY45soft;
  VLnY45soft.push_back(new TH1D("LnY45soft","Ln(1/y45) (Soft)",nBins,xLo,xHi));
  VLnY45soft[0]->SetMaximum(5.01); 
  VLnY45soft[0]->SetMinimum(0.00001); 
  VLnY45soft[0]->GetXaxis()->SetTitle("Ln(1/y45)");
  VLnY45soft[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VLnY45soft.push_back(new TH1D(string("LnY45soft_"+num2str(iWeight,1)).c_str()," ",nBins,xLo,xHi));  
  LiveDisplay dispLnY45soft(VLnY45soft,&pythia);
  dispLnY45soft.setDataFile(string(VINCIA_DAT+"Aleph-91-LnY45.dat").c_str(),
    10,30,40,50,60,70);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispLnY45soft.addHisFile(mcFileRead[iFile]+"-LnY45soft.dat");
  dispLnY45soft.normalize(true);
  dispLnY45soft.logY(true);

  nBins = 21;
  xLo = 3.8;
  xHi = 8.0;
  double wY56 = nBins/(xHi-xLo);
  vector<TH1D*> VLnY56;
  VLnY56.push_back(new TH1D("LnY56","Ln(1/y56) (Hard)",nBins, xLo, xHi));
  VLnY56[0]->SetMaximum(50.); 
  VLnY56[0]->SetMinimum(0.00001); 
  VLnY56[0]->GetXaxis()->SetTitle("Ln(1/y56)");
  VLnY56[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VLnY56.push_back(new TH1D(string("LnY56_"+num2str(iWeight,1)).c_str()," ",
                                          nBins,xLo,xHi));  
  LiveDisplay dispLnY56(VLnY56,&pythia);
  dispLnY56.setDataFile(string(VINCIA_DAT+"Aleph-91-LnY56.dat").c_str(),
    10,30,40,50,60,70);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispLnY56.addHisFile(mcFileRead[iFile]+"-LnY56.dat");
  dispLnY56.normalize(true);
  dispLnY56.logY(true);

  nBins = 15;
  xLo = 8.0;
  xHi = 11.0;
  double wY56soft = nBins/(xHi-xLo);
  vector<TH1D*> VLnY56soft;
  VLnY56soft.push_back(new TH1D("LnY56soft","Ln(1/y56) (Soft)",nBins,xLo,xHi));
  VLnY56soft[0]->SetMaximum(5.01); 
  VLnY56soft[0]->SetMinimum(0.0001); 
  VLnY56soft[0]->GetXaxis()->SetTitle("Ln(1/y56)");
  VLnY56soft[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++)  
      VLnY56soft.push_back(new TH1D(string("LnY56soft_"+num2str(iWeight,1)).c_str()," ",nBins,xLo,xHi));  
  LiveDisplay dispLnY56soft(VLnY56soft,&pythia);
  dispLnY56soft.setDataFile(string(VINCIA_DAT+"Aleph-91-LnY56.dat").c_str(),
    10,30,40,50,60,70);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispLnY56soft.addHisFile(mcFileRead[iFile]+"-LnY56soft.dat");
  dispLnY56soft.normalize(true);
  dispLnY56soft.logY(true);

  // BZ Angle
  nBins = 20;
  xLo   = 0.0;
  xHi   = 1.0;
  double wBZ = 20.0/1.0;
  vector<TH1D*> VChiBZ;
  VChiBZ.push_back(new TH1D("CosChiBZ","Bengtsson-Zerwas Angle",nBins,xLo,xHi));
  VChiBZ[0]->SetMaximum(4.0); 
  VChiBZ[0]->SetMinimum(-0.2); 
  VChiBZ[0]->GetXaxis()->SetTitle("cos(#Chi_{BZ})");
  VChiBZ[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) {
    for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VChiBZ.push_back(new TH1D(string("ChiBZ_"+num2str(iWeight,1)).c_str(),
          " ",nBins,xLo,xHi));
  }
  LiveDisplay dispChiBZ(VChiBZ,&pythia);
  dispChiBZ.setDataFile(string(VINCIA_DAT+"Delphi-91-chiBZ.dat").c_str(),
    10,30,40,50,60);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispChiBZ.addHisFile(mcFileRead[iFile]+"-ChiBZ.dat");
  dispChiBZ.normalize(true);

  // KSW Angle
  nBins = 20;
  xLo   = -1.0;
  xHi   =  1.0;
  double wKSW = 20.0/2.0;
  vector<TH1D*> VPhiKSW;
  VPhiKSW.push_back(new TH1D("CosPhiKSW","Korner-Schierholz-Willrodtorner Angle",nBins, xLo, xHi));
  VPhiKSW[0]->SetMaximum(1.5); 
  VPhiKSW[0]->SetMinimum(-0.075); 
  VPhiKSW[0]->GetXaxis()->SetTitle("cos(#Phi_{KSW})");
  VPhiKSW[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) {
    for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VPhiKSW.push_back(new TH1D(string("PhiKSW_"+num2str(iWeight,1)).c_str(),
          " ",nBins,xLo,xHi));
  }
  LiveDisplay dispPhiKSW(VPhiKSW,&pythia);
  dispPhiKSW.setDataFile(string(VINCIA_DAT+"Delphi-91-phiKSW.dat").c_str(),
    10,30,40,50,60);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispPhiKSW.addHisFile(mcFileRead[iFile]+"-PhiKSW.dat");
  dispPhiKSW.normalize(true);

  // NR Angle
  nBins = 20;
  xLo   =  0.0;
  xHi   =  1.0;
  double wNR = 20.0/1.0;
  vector<TH1D*> VThetaNR;
  VThetaNR.push_back(new TH1D("CosThetaNR","Nachtman-Reiter Angle", nBins, xLo, xHi));
  VThetaNR[0]->SetMaximum(2.5); 
  VThetaNR[0]->SetMinimum(-0.125); 
  VThetaNR[0]->GetXaxis()->SetTitle("|cos(#Theta_{NR})|");
  VThetaNR[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VThetaNR.push_back(new TH1D(string("ThetaNR_"+num2str(iWeight,1)).c_str(),
                                  " ",nBins,xLo,xHi));  
  LiveDisplay dispThetaNR(VThetaNR,&pythia);
  dispThetaNR.setDataFile(string(VINCIA_DAT+"Delphi-91-thetaNR.dat").c_str(),
    10,30,40,50,60);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispThetaNR.addHisFile(mcFileRead[iFile]+"-ThetaNR.dat");
  dispThetaNR.normalize(true);

  // 34 Angle
  nBins = 20;
  xLo   = -1.0;
  xHi   = 1.0;
  double wA34 = 20.0/2.0;  
  vector<TH1D*> VAlpha34;
  VAlpha34.push_back(new TH1D("CosAlpha34","#alpha_{34}", nBins, xLo, xHi));
  VAlpha34[0]->SetMaximum(1.2); 
  VAlpha34[0]->SetMinimum(-0.06); 
  VAlpha34[0]->GetXaxis()->SetTitle("cos(#alpha_{34})");
  VAlpha34[0]->GetYaxis()->SetTitle("Rate");
  if (vinciaOn) {
    for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VAlpha34.push_back(new TH1D(string("Alpha34_"+num2str(iWeight,1)).c_str(),
          " ",nBins,xLo,xHi));
  }
  LiveDisplay dispAlpha34(VAlpha34,&pythia);
  dispAlpha34.setDataFile(string(VINCIA_DAT+"Delphi-91-a34.dat").c_str(),
    10,30,40,50,60);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispAlpha34.addHisFile(mcFileRead[iFile]+"-Alpha34.dat");
  dispAlpha34.normalize(true);
  
  // R3bl 
  double nB(0), nC(0), nL(0);
  int nBinsR3bl = 44;
  xLo = 0.00;
  xHi = 0.055;
  double binWidR3bl = (xHi-xLo)/nBinsR3bl; 
  vector<TH1D*> VR3bl;
  VR3bl.push_back(new TH1D("R3bl","R3bl",nBinsR3bl, xLo, xHi));
  VR3bl[0]->SetMaximum(1.16); 
  VR3bl[0]->SetMinimum(0.85); 
  VR3bl[0]->GetXaxis()->SetTitle("y_{23}");
  VR3bl[0]->GetYaxis()->SetTitle("R3bl");
  if (vinciaOn) {
    for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VR3bl.push_back(new TH1D(string("R3bl_"+num2str(iWeight,1)).c_str(),
          " ",nBinsR3bl,xLo,xHi));
  }
  LiveDisplay dispR3bl(VR3bl,&pythia);
  dispR3bl.setDataFile(
    string(VINCIA_DAT+"Delphi-91-R3bl-Durham-had.dat").c_str(),20,40,50,60);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispR3bl.addHisFile(mcFileRead[iFile]+"-R3bl.dat");
  dispR3bl.normalize(false);
  dispR3bl.logY(false);
  dispR3bl.ratioRange(0.875,1.125);

  // nCharged
  TH1D *HnCharged=new TH1D("nCh","nCharged", 60, -0.5, 59.5);
  HnCharged->SetMinimum(0.00001);
  HnCharged->SetMaximum(0.5);

  // Ln(1/x)
  TH1D *Hx=new TH1D("LnXInv","Ln(1/x)", 200, 0.0, 5.4);
  double wx = 200.0/5.4;
  Hx->SetMinimum(0.0002);
  Hx->SetMaximum(8.);

  // Ratios of neutral 3rd jets with rapidity gaps for ALEPH CR study
  xLo = 0.50;
  xHi = 2.50;
  nBins = 2;
  vector<TH1D*> VrCRaleph;
  VrCRaleph.push_back(new TH1D("rCRaleph","ALEPH Gluon-Jet CR Observables",nBins, xLo, xHi));
  VrCRaleph[0]->SetMaximum(0.65); 
  VrCRaleph[0]->SetMinimum(-0.125); 
  VrCRaleph[0]->GetXaxis()->SetTitle("");
  VrCRaleph[0]->GetYaxis()->SetTitle("Value");
  VrCRaleph[0]->GetXaxis()->SetBinLabel(1,"r_{0}=#frac{N_{3j,gap,neut}}{N_{3j}}");
  VrCRaleph[0]->GetXaxis()->SetBinLabel(2,"f_{0}=#frac{N_{3j,gap,neut}}{N_{3j,gap}}");
if (vinciaOn) for (int iWeight=1; iWeight<nWeights; iWeight++) 
      VrCRaleph.push_back(new TH1D(string("rCRaleph_"+num2str(iWeight,1)).c_str(),
                                            " ",nBins,xLo,xHi));  
  LiveDisplay dispRCRaleph(VrCRaleph,&pythia);
  dispRCRaleph.setDataFile(string(VINCIA_DAT+"Aleph-91-cr3j.dat").c_str(),
		    20,0,40,50,60);
  for (int iFile=0; iFile<int(mcFileRead.size()); ++iFile)
    dispRCRaleph.addHisFile(mcFileRead[iFile]+"-rCRaleph.dat");
  dispRCRaleph.normalize(false);
  dispRCRaleph.logY(false);
  dispRCRaleph.ratioRange(0.5,1.5);

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settingsPtr->mode("Random:seed");
  vincia.rndmPtr->init(initseed);

  // Counter for negative-weight events
  double evwt=1.0;
  double sumw[nWeights];
  for (int iWeight=0; iWeight<nWeights; iWeight++) sumw[iWeight]=0.0;

  // Counters for jet rates
  nBins = 100;
  double sumN1j[nBins][nWeights];  
  double sumN2j[nBins][nWeights];
  double sumN3j[nBins][nWeights];
  double sumN4j[nBins][nWeights];
  double sumN5j[nBins][nWeights];
  double sumN6j[nBins][nWeights];
  double sumY23l[nBinsR3bl][nWeights];
  double sumY23b[nBinsR3bl][nWeights];
  for (int iBin=0; iBin<nBins; iBin++) 
    for (int iWeight=0; iWeight<nWeights; iWeight++) {
      sumN1j[iBin][iWeight] = 0.0;
      sumN2j[iBin][iWeight] = 0.0;
      sumN3j[iBin][iWeight] = 0.0;
      sumN4j[iBin][iWeight] = 0.0;
      sumN5j[iBin][iWeight] = 0.0;
      sumN6j[iBin][iWeight] = 0.0;
      if (iBin < nBinsR3bl) {
        sumY23l[iBin][iWeight] = 0.0;
        sumY23b[iBin][iWeight] = 0.0;
      }
    }

  // Counter for number of 3-jet events with yCut = 0.02 for ALEPH CR study
  double sumN3jYcut02           = 0.0;
  double sumN3jYcut02gap        = 0.0;
  double sumN3jYcut02neutralGap = 0.0;
  double sumN3jYcut02neutral    = 0.0;
  // FastJet needs bool to tell whether this event is the first one
  bool firstEvent = true; 

  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    if (!pythia.next()) continue;

    // Event weight
    if (vinciaOn) evwt    = vincia.weight();
    for (int iWeight=0; iWeight<nWeights; iWeight++) 
      sumw[iWeight] += vincia.weight(iWeight);

    // Count FS charged hadrons, partons, and quarks
    bool isZ2B   = (event[6].idAbs() == 5);
    bool isZ2C   = (event[6].idAbs() == 4);
    int  nCh     = 0;

    // Add final-state particles to FastJet input vector    
    fjInputs.resize(0);
    for (int i=1;i<event.size();i++) {

      // Final state only
      if (!pythia.event[i].isFinal())        continue;
      
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].status() > 80) {
        nCh++;
        double LnXInv=log(double(1.0/(2.0*event[i].pAbs()/eCM)));
        Hx->Fill(LnXInv,evwt*wx);
      }

      // Now decide what to pass to FastJet
      // NOTE: neutrinos are *included* in this list, since ALEPH
      // appears to have corrected their data to include them.
      // Store as input to Fastjet
      fastjet::PseudoJet particle(pythia.event[i].px(),
				  pythia.event[i].py(), 
				  pythia.event[i].pz(),
				  pythia.event[i].e() );
      particle.set_user_index(i);
      fjInputs.push_back( particle);      
    }
    
    // Counter for B events
    if (isZ2B) nB++;
    else if (isZ2C) nC++;
    else nL++;

    // FastJet
    if (fjInputs.size() == 0) {
      cout << "Error: event with no final state particles" << endl;
      continue;
    }

    // Run Fastjet algorithm    
    fastjet::ClusterSequence clustSeq(fjInputs, *jetDef);

    // 4-jet angles Use yCut from LEP 4-jet angle analysis
    vector<fastjet::PseudoJet> fastjets =  clustSeq.exclusive_jets_ycut(0.008);
    vector<fastjet::PseudoJet> sorted = fastjet::sorted_by_E(fastjets);

    // Extract merging scales
    double y56   = clustSeq.exclusive_ymerge(5);
    double lny56 = log(y56);
    double y45   = clustSeq.exclusive_ymerge(4);
    double lny45 = log(y45);
    double y34   = clustSeq.exclusive_ymerge(3);
    double lny34 = log(y34);
    double y23   = clustSeq.exclusive_ymerge(2);
    double lny23 = log(y23);
    double lny12 = log(clustSeq.exclusive_ymerge(1));

    // Fill Jet resolution histograms with results from FastJet
    for (int iWeight=0; iWeight < nWeights; iWeight++) { 
      if (iWeight != 0 && iWeight != 9) continue;
      VLnY12[iWeight]->Fill(-lny12,wY12*vincia.weight(iWeight));
      VLnY23[iWeight]->Fill(-lny23,wY23*vincia.weight(iWeight));
      VLnY23soft[iWeight]->Fill(-lny23,wY23soft*vincia.weight(iWeight));
      VLnY34[iWeight]->Fill(-lny34,wY34*vincia.weight(iWeight));
      VLnY34soft[iWeight]->Fill(-lny34,wY34soft*vincia.weight(iWeight));
      VLnY45[iWeight]->Fill(-lny45,wY45*vincia.weight(iWeight));
      VLnY45soft[iWeight]->Fill(-lny45,wY45soft*vincia.weight(iWeight));
      VLnY56[iWeight]->Fill(-lny56,wY56*vincia.weight(iWeight));
      VLnY56soft[iWeight]->Fill(-lny56,wY56soft*vincia.weight(iWeight));
      // R3bl (used in paper on massive antennae)
      int iy = -1;
      for (double y=0.0+binWidR3bl/2.; y<=0.055; y += binWidR3bl) {
        ++iy;
        // Check if event looks like 3-jet event at this scale
        if (y23 >= y && iy < nBinsR3bl) {
          if (isZ2B) sumY23b[iy][iWeight] += vincia.weight(iWeight);
          else if (!isZ2C) sumY23l[iy][iWeight] += vincia.weight(iWeight);
        }
      }
      VR3bl[iWeight]->Reset();
      iy = -1;
      for (double y=0.0+binWidR3bl/2.; y<=0.055; y += binWidR3bl) {
        ++iy;
        double valb  = sumY23b[iy][iWeight] / max(1.0,nB); 
        double errb  = sqrt(sumY23b[iy][iWeight]) / max(1.0,nB);
        double vall  = sumY23l[iy][iWeight] / max(1.0,nL);
        double errl  = sqrt(sumY23l[iy][iWeight]) / max(1.0,nL);
        double R3bl = 1.0;
        double errR = 0.0;
        if (vall > 0 && valb > 0) {
          R3bl = valb/vall;
          errR = sqrt(pow2(errb/valb) + pow2(errl/vall)) * R3bl;
        }
         int iBin    = VR3bl[iWeight]->FindBin(y);
        // Don't fill very first bin where there is no data (restricts y axis)
        if (iBin >= 3) {
          VR3bl[iWeight]->SetBinContent(iBin,R3bl);        
          VR3bl[iWeight]->SetBinError(iBin,errR);        
        }
      }

    }

    // Histogram jet rates
    // Loop over resolutions
    for (int iWeight=0; iWeight < nWeights; iWeight++) {
      if (iWeight != 0 && iWeight != 9) continue;
      Vr1j[iWeight]->Reset();
      Vr2j[iWeight]->Reset();
      Vr2jSoft[iWeight]->Reset();
      Vr3j[iWeight]->Reset();
      Vr3jSoft[iWeight]->Reset();
      Vr4j[iWeight]->Reset();
      Vr5j[iWeight]->Reset();
      Vr4jSoft[iWeight]->Reset();
      Vr5jSoft[iWeight]->Reset();
      Vr6j[iWeight]->Reset();
      int iy=-1;
      for (double lny=-10.1; lny<0.0; lny += 0.2) {
        iy++;
        if (lny12 < lny) sumN1j[iy][iWeight] += vincia.weight(iWeight);
        else if (lny23 < lny) sumN2j[iy][iWeight] += vincia.weight(iWeight);
        else if (lny34 < lny) sumN3j[iy][iWeight] += vincia.weight(iWeight);
        else if (lny45 < lny) sumN4j[iy][iWeight] += vincia.weight(iWeight);
        else if (lny56 < lny) sumN5j[iy][iWeight] += vincia.weight(iWeight);
        else  sumN6j[iy][iWeight] += vincia.weight(iWeight);
        double err, val;
        // 1j
        val = sumN1j[iy][iWeight]/sumw[iWeight];
        err = sqrt(sumN1j[iy][iWeight])/sumw[iWeight];
        int iBin = Vr1j[iWeight]->FindBin(lny);
        Vr1j[iWeight]->SetBinContent(iBin,val);
        Vr1j[iWeight]->SetBinError(iBin,err);
        // 2j
        Vr2j[iWeight]->Fill(lny,sumN2j[iy][iWeight]/sumw[iWeight]);
        err = sqrt(sumN2j[iy][iWeight])/sumw[iWeight];
        iBin = Vr2j[iWeight]->FindBin(lny);
        Vr2j[iWeight]->SetBinError(iBin,err);
        // 2j soft part
        Vr2jSoft[iWeight]->Fill(lny,sumN2j[iy][iWeight]/sumw[iWeight]);
        iBin = Vr2jSoft[iWeight]->FindBin(lny);
        Vr2jSoft[iWeight]->SetBinError(iBin,err);
        // 3j
        Vr3j[iWeight]->Fill(lny,sumN3j[iy][iWeight]/sumw[iWeight]);
        err = sqrt(sumN3j[iy][iWeight])/sumw[iWeight];
        iBin = Vr3j[iWeight]->FindBin(lny);
        Vr3j[iWeight]->SetBinError(iBin,err);
        // 3j soft part
        Vr3jSoft[iWeight]->Fill(lny,sumN3j[iy][iWeight]/sumw[iWeight]);
        iBin = Vr3jSoft[iWeight]->FindBin(lny);
        Vr3jSoft[iWeight]->SetBinError(iBin,err);
        // 4j
        Vr4j[iWeight]->Fill(lny,sumN4j[iy][iWeight]/sumw[iWeight]);
        err = sqrt(sumN4j[iy][iWeight])/sumw[iWeight];
        iBin = Vr4j[iWeight]->FindBin(lny);
        Vr4j[iWeight]->SetBinError(iBin,err);
        // 4j soft part
        Vr4jSoft[iWeight]->Fill(lny,sumN4j[iy][iWeight]/sumw[iWeight]);
        iBin = Vr4jSoft[iWeight]->FindBin(lny);
        Vr4jSoft[iWeight]->SetBinError(iBin,err);
        // 5j
        Vr5j[iWeight]->Fill(lny,sumN5j[iy][iWeight]/sumw[iWeight]);
        err = sqrt(sumN5j[iy][iWeight])/sumw[iWeight];
        iBin = Vr5j[iWeight]->FindBin(lny);
        Vr5j[iWeight]->SetBinError(iBin,err);
        // 5j soft part
        Vr5jSoft[iWeight]->Fill(lny,sumN5j[iy][iWeight]/sumw[iWeight]);
        iBin = Vr5jSoft[iWeight]->FindBin(lny);
        Vr5jSoft[iWeight]->SetBinError(iBin,err);
        // 6j
        Vr6j[iWeight]->Fill(lny,sumN6j[iy][iWeight]/sumw[iWeight]);
        err = sqrt(sumN6j[iy][iWeight])/sumw[iWeight];
        iBin = Vr6j[iWeight]->FindBin(lny);
        Vr6j[iWeight]->SetBinError(iBin,err);
      }
    }

    // Fill charged-particle histogram
    HnCharged->Fill(1.0*nCh,evwt);

    // Special analysis of 3-jet events, for CR constraints
    // Define 3-jet event as having y34 < 0.02 (from hep-ex/0604042)
    // The calculation at the particle level of the event generator (’MC,gen’) is done separately according to the following prescription. Hadronic Z events are generated without initial- and final-state photon radiation. The Durham cluster algorithm is applied to all charged and neutral stable particles to select three-jet events. The jets are ordered according to their actual energies. Rapidities are calculated using the actual particle masses. To define a rapidity gap, neutral particles with energies smaller than 0.6 GeV are omitted.
    // Vincia doesn't have QED radiation, so that's fine but need to check how much
    // it may affect Pythia comparisons. We can always choose to veto events in which
    // a photon appears as an FSR particle.
    bool isNeutral = true;
    if (y34 < 0.02 && sorted.size() >= 3) {
      sumN3jYcut02++;
      fastjet::PseudoJet jet3 = sorted[2];
      // Get the constituents and loop over them
      double sumQ = 0.;
      vector<fastjet::PseudoJet> constituents = jet3.constituents();
      // Find Jet axis and charge sum
      Vec4 jetAxis;
      for (int iC=0; iC<(int)constituents.size(); ++iC) {
	int i = constituents[iC].user_index();
	double qC = event[i].charge();
	sumQ += qC;
	jetAxis += event[i].p();
      }
      if (abs(sumQ) > 0.1) isNeutral = false;
      // Second loop: compute rapidities of contituents along jet axis
      // to define rapidity gap
      bool hasRapGap = true;
      for (int iC=0; iC<(int)constituents.size(); ++iC) {
	// Retreive Pythia event index from Fastjet list
	int i         = constituents[iC].user_index();
	// Discard neutral particles with E < 0.6 GeV
	double ei     = event[i].e();
	double qCtype = event[i].chargeType();
	if (qCtype == 0 && ei < 0.6) continue;	
	// Compute rapidity along jet axis
	Vec4 pi       = event[i].p(); 
	double cosTheta = dot3(pi,jetAxis)/pi.pAbs()/jetAxis.pAbs();
	double pL     = pi.pAbs()*cosTheta;
	double pT2    = pow2(pi.pAbs())*(1-pow2(cosTheta));
	double yTrue  = log( (ei + pL) / sqrt(event[i].m2() + pT2) );	
	// Check for rapidity gap in region (0,1.5)
	if (0 < yTrue && yTrue < 1.5) hasRapGap = false;
      }
      // Neutral jets; allow for (small) rounding errors
      if (isNeutral && hasRapGap) ++sumN3jYcut02neutralGap;
      if (isNeutral) ++sumN3jYcut02neutral;
      if (hasRapGap) ++sumN3jYcut02gap;
      // Update histogram
      for (int iWeight=0; iWeight < nWeights; iWeight++) {
	VrCRaleph[iWeight]->SetBinContent(1,sumN3jYcut02neutralGap/sumN3jYcut02);
	VrCRaleph[iWeight]->SetBinError(1,sqrt(sumN3jYcut02neutralGap)/sumN3jYcut02);
	if (sumN3jYcut02gap>0) {
	  VrCRaleph[iWeight]->SetBinContent(2,sumN3jYcut02neutralGap/sumN3jYcut02gap);
	  VrCRaleph[iWeight]->SetBinError(2,2*sqrt(sumN3jYcut02neutralGap)/sumN3jYcut02gap);
	}
      }
    }
    
    // Call root app updater
    liveDisplaysApp.update(); 

    if ( iEvent >= iSave ) {
      iSave       = liveDisplaysApp.nNextSave();
      cout<<" At event = "<< iEvent <<"  (next save at "<<iSave<<")"<<endl;
    }
    
    // Now only keep exclusive 4-jet events
    if (sorted.size() != 4) continue;

    // For the first event, print the FastJet details
    if (firstEvent) {
      cout << "Ran " << jetDef->description() << endl;
      cout << "Strategy adopted by FastJet was "
           << clustSeq.strategy_string() << endl << endl;      
      firstEvent = false;
    }

    // Compute 4-jet angles
    Vec4 p1(sorted[0].px(),sorted[0].py(),sorted[0].pz(),sorted[0].e());
    Vec4 p2(sorted[1].px(),sorted[1].py(),sorted[1].pz(),sorted[1].e());
    Vec4 p3(sorted[2].px(),sorted[2].py(),sorted[2].pz(),sorted[2].e());
    Vec4 p4(sorted[3].px(),sorted[3].py(),sorted[3].pz(),sorted[3].e());
    // Bengtsson-Zerwas
    Vec4 v1 = cross3(p1,p2);
    Vec4 v2 = cross3(p3,p4);
    double cosChiBZ = abs(costheta(v1,v2));
    // KSW
    v1 = cross3(p1,p4);
    v2 = cross3(p2,p3);
    double alpha1 = theta(v1,v2);
    v1 = cross3(p1,p3);
    v2 = cross3(p2,p4);
    double alpha2 = theta(v1,v2);
    double cosPhiKSW = cos((alpha1+alpha2)/2.);
    // Nachtman-Reiter
    v1 = p1 - p2;
    v2 = p3 - p4;
    double cosThetaNR = abs(costheta(v1,v2));
    // Alpha34
    double cosA34 = costheta(p3,p4);

    // Fill 4-jet angle histograms
    for (int iWeight=0; iWeight < nWeights; iWeight++) {
      VChiBZ[iWeight]->Fill(cosChiBZ,vincia.weight(iWeight)*wBZ);
      VThetaNR[iWeight]->Fill(cosThetaNR,vincia.weight(iWeight)*wNR);
      VAlpha34[iWeight]->Fill(cosA34,vincia.weight(iWeight)*wA34);
      VPhiKSW[iWeight]->Fill(cosPhiKSW,vincia.weight(iWeight)*wKSW);
    }
  }

  // End-of-run forced display update and saves
  liveDisplaysApp.update(true);

  // Print out Vincia internal histograms
  if (vinciaOn && verbose >= 2) vincia.printHistos();

  // Print out Vincia end-of-run information
  pythia.stat();
  if (vinciaOn) vincia.printInfo();

  // Example of saving histograms to ROOT file
  cout<<"Saving root file... "<<outfile.str()<<endl;
  TFile f(outfile.str().c_str(),"recreate");
  for (int iWeight=0; iWeight<nWeights; ++iWeight) {
    VLnY12[iWeight]->Write();
    VLnY23[iWeight]->Write();
    VLnY34[iWeight]->Write();
    VLnY45[iWeight]->Write();
    VLnY56[iWeight]->Write();
    VChiBZ[iWeight]->Write();
    VThetaNR[iWeight]->Write();
    VAlpha34[iWeight]->Write();
    VPhiKSW[iWeight]->Write();
  }
  f.Close();
  
  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
