// LiveDisplays.cc is a part of the LiveDisplays plug-in to PYTHIA 8
// Copyright (C) 2019 P. Skands.
// LiveDisplays and Pythia 8 are licenced under the GNU GPL v2 or later.
// Please respect the MCnet Guidelines for Event Generator Authors and Users,
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the LiveDisplay class

// Include the LiveDisplays header
#include "LiveDisplays/LiveDisplays.h"

namespace Pythia8 {

//==========================================================================

// LiveDisplaysApp class member functions

//--------------------------------------------------------------------------

bool LiveDisplaysApp::update(bool forceUpdate, bool forceSave) {
  gSystem->ProcessEvents();
  bool doUpdate = (pythiaPtr->info.nAccepted() >= nNextUpdate());
  bool doSave = (pythiaPtr->info.nAccepted() >= nNextSave());
  if (doUpdate || doSave || forceUpdate || forceSave) {
    // Go through and update (and save) all canvases
    TSeqCollection* canvases = gROOT->GetListOfCanvases();
    TIter it(canvases);
    for (int i=0; i<canvases->GetSize(); ++i) {
      LiveDisplay* dispPtr = (LiveDisplay*)it.Next();
      if (dispPtr == 0) continue;
      dispPtr->update();
      if (doSave || forceSave) {
        dispPtr->saveLiveDisplay();
        dispPtr->saveHistogram();
      }
    }
    if (doUpdate) {
      // Update nNextUpdate
      nNextUpdateSav += deltaUpdate;
    }
    if (doSave) {
      nNextSaveSav = max(nNextSaveSav + deltaSave,
        nNextSaveSav * pow(2,1./fSave));
    }
    // Go through and update

    return true;
  }
  return false;
}

//==========================================================================

// LiveDisplay class member functions

//--------------------------------------------------------------------------

// Initialise

void LiveDisplay::init(Pythia* pythiaPtr) {
  // Convenient pointer to Pythia Settings database
  Settings* settingsPtr = &pythiaPtr->settings;
  // Main histogram display settings
  logYsav          = settingsPtr->flag("LiveDisplays:logY");
  logXsav          = settingsPtr->flag("LiveDisplays:logX");
  normalizeSav     = settingsPtr->flag("LiveDisplays:normalize");
  normFacSav       = settingsPtr->parm("LiveDisplays:normFactor");
  rebinLikeDataSav = settingsPtr->flag("LiveDisplays:rebinLikeData");
  zeroSkippingSav  = settingsPtr->flag("LiveDisplays:zeroSkipping");
  // Ratio pad settings
  showRatio(settingsPtr->flag("LiveDisplays:showRatio"));
  ratioMinSav = settingsPtr->parm("LiveDisplays:ratioMin");
  ratioMaxSav = 1.+0.999*(settingsPtr->parm("LiveDisplays:ratioMax")-1);
  ratioLogYsav = settingsPtr->flag("LiveDisplays:ratioLogY");
  // Relative uncertainty pad settings
  showRelUnc(settingsPtr->flag("LiveDisplays:showRelUnc"));
  // Legend settings
  vector<string> legendTmp;
  legendTmp.push_back(settingsPtr->word("LiveDisplays:legendLabel"));
  setLegend(legendTmp);
  showChi2(settingsPtr->flag("LiveDisplays:showChi2"));
  forceLeftLegend(false);
  // Main Plot Title
  titleSizeSav  = settingsPtr->parm("LiveDisplays:titleSize");
  // Subtitle 1 & 2 settings
  subTitle1(settingsPtr->word("LiveDisplays:SubTitle1"));
  subTitle1ySav    = settingsPtr->parm("LiveDisplays:SubTitle1y");
  subTitle1sizeSav = settingsPtr->parm("LiveDisplays:SubTitle1size");
  subTitle2(settingsPtr->word("LiveDisplays:SubTitle2"));
  subTitle2ySav    = settingsPtr->parm("LiveDisplays:SubTitle2y");
  subTitle2sizeSav = settingsPtr->parm("LiveDisplays:SubTitle2size");
  // Top left and right labels
  leftLabelSav  = settingsPtr->word("LiveDisplays:leftLabel");
  rightLabelSav = settingsPtr->word("LiveDisplays:rightLabel");
  hasLeftLabel  = (leftLabelSav != "none");
  hasRightLabel = (rightLabelSav != "none");
  // Save files
  saveFilePrefixSav = settingsPtr->word("LiveDisplays:saveFilePrefix");
  saveFileSuffixSav = settingsPtr->word("LiveDisplays:saveFileSuffix");
  // Save graphics
  savePDFsav  = settingsPtr->flag("LiveDisplays:savePDF");
  saveEPSsav  = settingsPtr->flag("LiveDisplays:saveEPS");
  //Save to bitmapped files does not appear to work properly in ROOT.
  savePNGsav  = false;
  saveJPEGsav = false;
  saveROOTsav = settingsPtr->flag("LiveDisplays:saveROOT");
  saveTEXsav  = settingsPtr->flag("LiveDisplays:saveTEX");
  // Uncertainty band labels;
  uLabelsSav.clear();
  for (int i=1; i<=pythiaPtr->info.nWeights(); ++i)
    uLabelsSav.push_back(pythiaPtr->info.weightLabel(i));
  // Non-production parameters
  titleSav   = "default";
  optStatSav = settingsPtr->mode("LiveDisplays:optStat");
  verboseSav = settingsPtr->flag("LiveDisplays:verbose");

}

//--------------------------------------------------------------------------

// Add pad (for ratio to data or decomposition of vector of histos)

void LiveDisplay::addPad() {
  nPad++;
}

//--------------------------------------------------------------------------

// Goto pad (or create new if not yet existing)

void LiveDisplay::gotoPad(int iPadIn) {
  cd();

  // If pads already defined, just cd to the desired one
  if (isInitPad) {
    gPad->cd(min(iPadIn,nPad));
    return;
  }

  // If pads not yet defined, create them
  gPad->Divide(1,nPad);
  double nDiv = 4. + nPad;
  double topMargin    = 0.056;
  double bottomMargin = 0.09;
  if (nPad >= 2) bottomMargin *= 1.75 + 0.5*nPad;
  double rightMargin  = 0.045;
  double leftMargin   = 0.135;
  if (nPad >= 2) leftMargin += 0.01;
  // Left margin at least 35 pixels
  double leftMarginPix = max(35.,leftMargin * gPad->GetWw());
  leftMargin = leftMarginPix/gPad->GetWw();
  // Left margin is at least 1.5 times right one (for symmetric look)
  leftMargin = max(leftMargin,1.5*rightMargin);
  for (int iPad=1; iPad<=nPad; ++iPad) {
    cd();
    gPad->cd(iPad);
    double yHi = 1.0;
    if (iPad >= 2) yHi = 1.0 - double(2+iPad)/nDiv;
    double yLo = 1.0 - double( (iPad+3)/nDiv );
    if (iPad == nPad) yLo = 0.0;
    // Set pad coordinates
    gPad->SetPad(0.0,yLo,1.0,yHi);
    // Set inner margins : top margin
    if (iPad == 1) gPad->SetTopMargin(topMargin);
    else gPad->SetTopMargin(0.0);
    // Set inner margins : bottom margin
    if (iPad == 1 && nPad >= 2) gPad->SetBottomMargin(0.03);
    else if (iPad == nPad) gPad->SetBottomMargin(bottomMargin);
    else gPad->SetBottomMargin(0.0);
    gPad->SetLeftMargin(leftMargin);
    gPad->SetRightMargin(rightMargin);
    // Optionally set log axes
    if (logXsav) gPad->SetLogx();

    if (!has2d) {
      // Modify labels
      // When adding many pads, increase label font sizes
      if (iPad >= 2) {
        double labSizeX = 1.13 * axesPtr->GetXaxis()->GetLabelSize();
        double labSizeY = 1.13 * axesPtr->GetYaxis()->GetLabelSize();
        axesPtr->GetXaxis()->SetLabelSize(labSizeX);
        axesPtr->GetYaxis()->SetLabelSize(labSizeY);
        double titSizeX = 1.16 * axesPtr->GetXaxis()->GetTitleSize();
        double titSizeY = 1.16 * axesPtr->GetYaxis()->GetTitleSize();
        axesPtr->GetXaxis()->SetTitleSize(titSizeX);
        axesPtr->GetYaxis()->SetTitleSize(titSizeY);
        double titOffsetY = 0.9*axesPtr->GetYaxis()->GetTitleOffset();
        axesPtr->GetYaxis()->SetTitleOffset(titOffsetY);
        // Ratio pad
        double labNormSizeX = 1.09 * axesNormPtr->GetXaxis()->GetLabelSize();
        double labNormSizeY = 1.09 * axesNormPtr->GetYaxis()->GetLabelSize();
        axesNormPtr->GetXaxis()->SetLabelSize(labNormSizeX);
        axesNormPtr->GetYaxis()->SetLabelSize(labNormSizeY);
        double titNormSizeX = 1.2*axesNormPtr->GetXaxis()->GetTitleSize();
        double titNormSizeY = 1.2*axesNormPtr->GetYaxis()->GetTitleSize();
        axesNormPtr->GetXaxis()->SetTitleSize(titNormSizeX);
        axesNormPtr->GetYaxis()->SetTitleSize(titNormSizeY);
        axesNormPtr->GetYaxis()->SetTitleOffset(0.5*titOffsetY);
      }
      // if there is more than one pad, don't display labels and titles on
      // x axis of first pad
      if ( (iPad >= 2) && (axesPtr) ) {
        axesPtr->GetXaxis()->SetLabelOffset(1.);
        axesPtr->GetXaxis()->SetTitleOffset(1.);
      }
      // Draw label closer if we have more pads
      if ( nPad >= 3) {
        axesNormPtr->GetYaxis()->SetTitleOffset(
          0.9*axesNormPtr->GetYaxis()->GetTitleOffset());
      }
      // RelUnc pad(s)
      if ( (nPad >= 3) && (axesVarPtr) ) {
        double labVarSizeX = 1.75 * axesVarPtr->GetXaxis()->GetLabelSize();
        double labVarSizeY = 1.3 * axesVarPtr->GetYaxis()->GetLabelSize();
        axesVarPtr->GetXaxis()->SetLabelSize(labVarSizeX);
        axesVarPtr->GetYaxis()->SetLabelSize(labVarSizeY);
        double titVarSizeX = max(1.75 * axesVarPtr->GetXaxis()->GetTitleSize(),
          double(axesVarPtr->GetXaxis()->GetLabelSize()));
        double titVarSizeY = 1.3 * axesVarPtr->GetYaxis()->GetTitleSize();
        double titVarOffsetY = 0.7 * axesVarPtr->GetYaxis()->GetTitleOffset();
        axesVarPtr->GetXaxis()->SetTitleSize(titVarSizeX);
        axesVarPtr->GetYaxis()->SetTitleSize(titVarSizeY);
        axesVarPtr->GetYaxis()->SetTitleOffset(titVarOffsetY);
        isInitPad = false;
      }
    }

    // Draw pad
    if (!isInitPad) {
      gPad->Draw();
    } else {
      gPad->Modified();
    }
  }
  isInitPad = true;
  gPad->cd(min(iPadIn,nPad));
}

//--------------------------------------------------------------------------

// Check if axes and/or data has been defined
// Modifications to single-pane displays should be done here
// Modifications specific to multi-pane displays should be done above

void LiveDisplay::setAxes(TH1D* initHis) {
  if (hasAxes()) {
    delete axesPtr;
    delete xCtrPtr;
    delete xUppPtr;
    delete axesNormPtr;
    delete axesVarPtr;
  }
  cd();
  gPad->cd(1);
  axesPtr = new TH1D(*initHis);
  axesPtr->Reset("ICES");
  hasAxesSav = true;
  // Axis histogram is pure dummy, has no stats, so switch off stats display
  axesPtr->SetStats(false);
  // Set fonts and axis tick divisions
  axesPtr->GetXaxis()->SetLabelFont(42);
  axesPtr->GetYaxis()->SetLabelFont(42);
  axesPtr->GetXaxis()->SetNdivisions(505);
  axesPtr->GetYaxis()->SetNdivisions(606);
  axesPtr->GetXaxis()->SetTitleFont(42);
  double labSizeX = 1.2 * axesPtr->GetXaxis()->GetLabelSize();
  double labSizeY = 1.2 * axesPtr->GetYaxis()->GetLabelSize();
  axesPtr->GetXaxis()->SetLabelSize(labSizeX);
  axesPtr->GetYaxis()->SetLabelSize(labSizeY);
  axesPtr->GetYaxis()->SetTitleSize(1.1*axesPtr->GetYaxis()->GetTitleSize());
  double titOffset = 1.2*axesPtr->GetYaxis()->GetTitleOffset();
  // More space for extreme axis ranges (more digits)
  double axisMin = axesPtr->GetMinimum();
  double axisMax = axesPtr->GetMaximum();
  // Check if asked for log x axes and whether low = 0
  if (gPad->GetLogx() && axisMin == 0.0) {
    axisMin += 0.5*axesPtr->GetBinLowEdge(2);
    axesPtr->SetAxisRange(axisMin, axisMax);
  }
  if ( (min(abs(axisMin),abs(axisMax)) < 1e-10
      && min(abs(axisMin),abs(axisMax)) != 0.0)
    || max(abs(axisMin),abs(axisMax)) > 1e10) titOffset *= 1.2;
  else if ( (min(abs(axisMin),abs(axisMax)) < 1e-2
      && min(abs(axisMin),abs(axisMax)) != 0.0)
    || max(abs(axisMin),abs(axisMax)) > 1e2) titOffset *= 1.1;
  // Reduce/increase offset for extreme aspect ratios
  titOffset *= sqrt(gPad->GetWh()/(1.0*gPad->GetWw()));
  titOffset = min(1.6,titOffset);
  axesPtr->GetYaxis()->SetTitleOffset(titOffset);
  axesPtr->GetYaxis()->SetTitleFont(42);
  // Switch off lame ROOT title
  axesPtr->SetTitle("");

  // Set axis histo for ratio pad
  cd();
  gPad->cd(nPad);
  axesNormPtr = new TH1D(*axesPtr);
  axesNormPtr->SetMinimum(ratioMinSav);
  axesNormPtr->SetMaximum(ratioMaxSav);
  // X axis
  axesNormPtr->GetXaxis()->SetTickLength(0.1);
  axesNormPtr->GetXaxis()->SetLabelSize(
    1.9*axesPtr->GetXaxis()->GetLabelSize());
  axesNormPtr->GetXaxis()->SetTitleFont(42);
  axesNormPtr->GetXaxis()->SetTitleSize(
    1.85*axesPtr->GetXaxis()->GetTitleSize());
  axesNormPtr->GetXaxis()->SetTitleOffset(
    1.*axesPtr->GetXaxis()->GetTitleOffset());
  // Y axis
  axesNormPtr->GetYaxis()->SetLabelFont(42);
  axesNormPtr->GetYaxis()->SetLabelSize(
    1.8*axesPtr->GetYaxis()->GetLabelSize());
  axesNormPtr->GetYaxis()->SetLabelOffset(0.01);
  axesNormPtr->GetYaxis()->SetTitleFont(42);
  axesNormPtr->GetYaxis()->SetTitleSize(
    1.8*axesPtr->GetYaxis()->GetTitleSize());
  axesNormPtr->GetYaxis()->SetTitle("Theory/Data");
  // Protect against ridiculous aspect ratios

  axesNormPtr->GetYaxis()->CenterTitle();
  axesNormPtr->GetYaxis()->SetTickLength(0.025);
  // Make copy of x axis for overdrawing
  int nBin = axesNormPtr->GetNbinsX();
  xCtrPtr = new TLine(axesNormPtr->GetBinLowEdge(1),1.0,
    axesNormPtr->GetBinLowEdge(nBin+1),1.0);
  xUppPtr = new TLine(axesNormPtr->GetBinLowEdge(1),ratioMaxSav,
    axesNormPtr->GetBinLowEdge(nBin+1),ratioMaxSav);


  // Set axis histo for relUnc pads (uncertainty variations)
  cd();
  gPad->cd(min(nPad,2));
  axesVarPtr = new TH1D(*axesPtr);
  axesVarPtr->SetMinimum(-1.5);
  axesVarPtr->SetMaximum(1.5);
  // X axis
  axesVarPtr->GetXaxis()->SetTickLength(
    2.*axesPtr->GetXaxis()->GetTickLength());
  axesVarPtr->GetXaxis()->SetLabelOffset(
    2.*axesPtr->GetXaxis()->GetLabelOffset());
  axesVarPtr->GetXaxis()->SetLabelSize(
    2.*axesPtr->GetXaxis()->GetLabelSize());
  axesVarPtr->GetXaxis()->SetTitleFont(42);
  axesVarPtr->GetXaxis()->SetTitleSize(
    2.5*axesPtr->GetXaxis()->GetTitleSize());
  axesVarPtr->GetXaxis()->SetTitleOffset(
    1.*axesPtr->GetXaxis()->GetTitleOffset());
  // Y axis
  axesVarPtr->GetYaxis()->SetNdivisions(203);
  axesVarPtr->GetYaxis()->SetTickLength(axesPtr->GetYaxis()->GetTickLength());
  axesVarPtr->GetYaxis()->SetLabelFont(42);
  axesVarPtr->GetYaxis()->SetLabelSize(
    1.8*axesPtr->GetYaxis()->GetLabelSize());
  axesVarPtr->GetYaxis()->SetLabelOffset(
    1.8*axesPtr->GetYaxis()->GetLabelOffset());
  axesVarPtr->GetYaxis()->SetTitleFont(42);
  axesVarPtr->GetYaxis()->SetTitle("Rel. Var.");
  axesVarPtr->GetYaxis()->SetTitleSize(
    2*axesPtr->GetYaxis()->GetTitleSize());
  axesVarPtr->GetYaxis()->SetTitleOffset(
    0.47*axesPtr->GetYaxis()->GetTitleOffset());
  axesVarPtr->GetYaxis()->CenterTitle();
}

//--------------------------------------------------------------------------

// Set file to be treated as experimental data,
// with i1 : i19 specifying data file format, columnwise (see VINCIA docs)

bool LiveDisplay::setDataFile(string dataFileName, int i1, int i2, int i3,
  int i4, int i5, int i6, int i7, int i8, int i9, int i10, int i11, int i12,
  int i13, int i14, int i15, int i16, int i17, int i18, int i19) {

  // Verbose output
  if (verboseSav) cout<<"(LiveDisplay): adding dataFile "<<dataFileName
                      <<" to display "<<displayNameSav<<endl;

  // Create new dataFilePtr with specified format
  vector<int> format;
  getFormat(format,i1,i2,i3,i4,i5,i6,i7,i8,i9,i10,i11,i12,i13,i14,i15,i16,
    i17,i18,i19);
  dataFilePtr = new VRfile(format);

  // Read file
  bool ok = dataFilePtr->readFile(dataFileName, verboseSav);
  if (ok) {

    // Protect against memory leaks
    if (hasDataSav) {
      delete dataPtr;
      delete dataNormPtr;
      delete dataTwoSigmaPtr;
    }

    // Create new VRgraph to contain data points
    dataPtr = new VRgraphData(dataFilePtr);
    dataNormPtr = new VRgraphNorm(dataPtr,dataPtr);
    dataNormPtr->setColorScheme(0);
    // Copies to display two-sigma uncertainties
    dataTwoSigmaPtr = new VRgraphData(dataFilePtr);
    dataTwoSigmaPtr->setColorScheme(-2);
    dataTwoSigmaNormPtr = new VRgraphNorm(dataTwoSigmaPtr,dataTwoSigmaPtr);
    dataTwoSigmaNormPtr->setColorScheme(-2);

    // This display has data
    hasDataSav = true;
    if (showRatioSav && nPad <= 1) addPad();

    // Check if rebinning histograms to same binning as data
    if (rebinLikeDataSav) {
      for (int iGraph=0; iGraph<int(graphs.size()); ++iGraph)
        graphs[iGraph].setRebinAxis(dataPtr,zeroSkippingSav);
      for (int iGraph=0; iGraph<int(graphsHis.size()); ++iGraph)
        graphsHis[iGraph].setRebinAxis(dataPtr,zeroSkippingSav);
    }

  } else {
    cout<<"VRLiveDisplay::setDataFile() error reading "<<dataFileName<<endl;
  }

  // Return
  return ok;
}

//------------------------------------------------------------------------

// addHisFile() :
// Add file with saved histogram,
// hasOverFlow: true if first and last data file points represent over- and
// underflow bins
// with i1 : i19 specifying the data format, columnwise (see VINCIA docs)

bool LiveDisplay::addHisFile(string hisFileName, int i1, int i2, int i3, int i4,
  int i5, int i6, int i7, int i8, int i9, int i10, int i11, int i12, int i13,
  int i14, int i15, int i16, int i17, int i18, int i19) {

  // Verbose output
  if (verboseSav) cout<<"(LiveDisplay): adding histogram file "<<hisFileName
                      <<" to display "<<displayNameSav<<endl;

  // Add to list of hisFilePtrs.
  int nHisFiles = hisFilePtrs.size();
  hisFilePtrs.resize(nHisFiles+1);

  // Create new dataFilePtr with specified format
  vector<int> format;
  getFormat(format,i1,i2,i3,i4,i5,i6,i7,i8,i9,i10,i11,i12,i13,i14,i15,i16,i17,
    i18,i19);
  hisFilePtrs[nHisFiles] = new VRfile(format);

  // Read file and return success/failure
  bool ok = hisFilePtrs[nHisFiles]->readFile(hisFileName, verboseSav);
  if (ok) {

    // Add to list of graphsHis
    int nGraphsHis = graphsHis.size();
    graphsHis.resize(nGraphsHis+1);

    // Create new VRgraph to contain histogram points
    graphsHis[nGraphsHis] = VRgraphHis( hisFilePtrs[nHisFiles] );
    graphsHis[nGraphsHis].setColorScheme(graphs.size()+graphsHis.size());

    // Store name
    hisFileNames.resize(nGraphsHis+1);
    hisFileNames[nGraphsHis] = hisFileName;

    // Check if rebinning main histogram to data
    if (rebinLikeDataSav && hasDataSav && dataPtr != 0)
      graphsHis[nGraphsHis].setRebinAxis(dataPtr,zeroSkippingSav);

    // Recreate list of normalized histograms
    // (involves pointers to elements of a vector so must be explicit)
    graphsHisNorm.resize(0);
    for (int i=0; i<int(graphsHis.size()); ++i) {
      VRgraphHis* hisPtr = &graphsHis[i];
      if (hasDataSav) {
        graphsHisNorm.push_back( VRgraphNorm(hisPtr,dataPtr) );
      } else if (graphs.size() >= 1) {
        graphsHisNorm.push_back(VRgraphNorm(hisPtr,&graphs[0]));
      } else {
        graphsHisNorm.push_back(VRgraphNorm(hisPtr,&graphsHis[0]));
      }
      graphsHisNorm[i].setColorScheme(graphs.size()+i+1);
    }

    // If there is more than one histogram, we can have a ratio pad
    if (graphsHis.size() + graphs.size() >= 2
      && showRatioSav && nPad <= 1) addPad();

    // Add pad for decomposition of vector band
    if (nPad <= 2 && showRelUncSav) {
      for (int i=0; i<int(graphs.size()); ++i)
        if (graphs[i].hasOuter()) addPad();
    }

  } else {
    cout<<"VRLiveDisplay::addHisFile() error reading "<<hisFileName<<endl;
    delete hisFilePtrs[nHisFiles];
    hisFilePtrs.pop_back();
  }

  // Return
  return ok;

}

//------------------------------------------------------------------------

// Force update of histogram data files, useful when doing several
// runs in parallel.

bool LiveDisplay::refreshHisFiles() {
  bool ok = true;
  for (int i=0; i<(int)hisFileNames.size(); ++i)
    ok = ok && hisFilePtrs[i]->readFile(hisFileNames[i], verboseSav);
  return ok;
}

//------------------------------------------------------------------------

// update():
// Update LiveDisplay

void LiveDisplay::update() {

  // Refresh any histogram files in case they changed since last update
  refreshHisFiles();

  // CD to this display
  cd();

  // If several pads, first do main plot
  gotoPad(1);

  // First check if this is 2D, then plot and exit. Rest is 1D.
  if (has2d) {
    // Draw with colored contours
    th2dPtr->Draw("colz");
    // Draw lines
    if (linesPtr.size() >= 1) {
      for (unsigned int iLine=0; iLine < linesPtr.size(); ++iLine)
        linesPtr[iLine]->Draw();
    }
    // Then update canvas
    cd();
    Modified();
    Update();
    gSystem->ProcessEvents();
    return;
  }

  // Update and (re)draw axis
  if (!hasAxes()) {
    if (graphs.size() >= 1) setAxes(graphs[0].getTH1DPtr());
    else return;
  }

  // Update max/min if possible.
  if (graphs.size() >= 1) {
    TH1D* hisPtr = graphs[0].getTH1DPtr();
    if (hisPtr != NULL) {
      // Check if min/max changed by user
      double yMinSav = axesPtr->GetMinimum();
      double yMaxSav = axesPtr->GetMaximum();
      double yMinNow = hisPtr->GetMinimum();
      double yMaxNow = hisPtr->GetMaximum();
      if (yMinSav != yMinNow || yMaxSav != yMaxNow) {
        axesPtr->SetMinimum(yMinNow);
        axesPtr->SetMaximum(yMaxNow);
      }

      // Also update log Y axis properties (force linear if yMin <= 0.0)
      if ( yMinNow <= 0.0 && gPad->GetLogy() ) logY(false);
      else if ( yMinNow > 0.0 && logYsav && !gPad->GetLogy() ) logY(true);

      // Update title
      title(titleSav);
    }
  }

  // Draw axes
  if (nPad >= 2) {
    // Switch off X axis title for main histo
    axesPtr->GetXaxis()->SetTitle("");
    // Draw relUnc pads
    if (showRelUncSav) {
      int iPad = 1;
      for (int iGraph = 0; iGraph < int(graphs.size()); ++iGraph) {
        if (graphs[iGraph].hasOuter()) {
          ++iPad;
          cd(iPad);
          axesVarPtr->Draw();
        }
      }
    }
    // Draw ratio pad
    if (showRatioSav && hasDataSav && dataPtr) {
      cd(nPad);
      axesNormPtr->Draw();
    }
    else if (showRatioSav && nPad >= 2) {
      cd(nPad);
      axesNormPtr->Draw();
    }
  }
  // Draw main display
  cd(1);
  axesPtr->Draw();

  // Update data
  // (data is always shown with 1 sigma error bars)
  if (hasDataSav && dataPtr) {
    cd(1);
    dataTwoSigmaPtr->update(normalizeSav,normFacSav,2.0);
    dataTwoSigmaPtr->DrawOuter("E0Z");
    dataPtr->update(normalizeSav,normFacSav,1.0);
    dataPtr->DrawOuter("E10");
    dataPtr->Draw("PE1");
    if (dataPtr->auxGraphs.size() > 0) {
      map<string, VRgraph>::iterator it;
      for (it=dataPtr->auxGraphs.begin(); it!=dataPtr->auxGraphs.end(); ++it)
        it->second.Draw("P");
    }
    if (dataNormPtr && nPad >= 2 && showRatioSav) {
      cd(nPad);
      dataTwoSigmaNormPtr->update(true,ratioMinSav,ratioMaxSav);
      dataTwoSigmaNormPtr->DrawOuter("e2");
      dataTwoSigmaNormPtr->Draw("e2");
      dataNormPtr->update(true,ratioMinSav,ratioMaxSav);
      dataNormPtr->DrawOuter("e2");
      dataNormPtr->Draw("e2");
      axesNormPtr->Draw("sameaxis");
      //gPad->RedrawAxis();
      // Draw center line at 1
      xCtrPtr->Draw();
    }
  }

  // Update graphs made from saved histograms (dat files)
  cd(1);
  // Determine plot range
  double xMinSav = axesPtr->GetBinLowEdge(1);
  int nBins = axesPtr->GetNbinsX();
  double xMaxSav = axesPtr->GetBinLowEdge(nBins)
    + axesPtr->GetBinWidth(nBins);
  for (unsigned int iGraph=0; iGraph<graphsHis.size(); ++iGraph) {

    // Set X scatter
    double scatterX = 0.0;
    // 0.1*pow(-1.0,(iGraph+graphs.size())*1.0)
    //  * (iGraph+graphs.size())/(graphsHis.size()+graphs.size());

    // Skip if update fails
    if (!graphsHis[iGraph].update(normalizeSav,normFacSav,scatterX))
      continue;

    // Remove points outside plot region (to avoid lines going outside region)
    for (int iPoint = graphsHis[iGraph].GetN()-1; iPoint >= 0; --iPoint) {
      double x, y;
      graphsHis[iGraph].GetPoint(iPoint,x,y);
      if (x < xMinSav || x > xMaxSav) graphsHis[iGraph].RemovePoint(iPoint);
    }

    // Outer band
    if (graphsHis[iGraph].hasOuter()) {
      int fillStyle = 3354;
      if (iGraph == 1) fillStyle = 3345;
      else if (iGraph >= 2) fillStyle = 3307;
      graphsHis[iGraph].getOuter()->SetFillStyle(fillStyle);
    }

    // Draw graph
    graphsHis[iGraph].DrawOuter("20");
    graphsHis[iGraph].Draw("LPZ0");
  }

  // Update graphs
  cd(1);
  for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {

    // Set X scatter
    double scatterX = 0.0;
    // 0.1*pow(-1.0,iGraph*1.0) * iGraph/graphs.size();

    // Skip if update fails
    if (!graphs[iGraph].update(normalizeSav,normFacSav,scatterX)) {
      continue;
    }

    // Outer band
    if (graphs[iGraph].hasOuter()) {
      int fillStyle = 3354;
      if (iGraph == 1) fillStyle = 3345;
      else if (iGraph >= 2) fillStyle = 3307;
      graphs[iGraph].getOuter()->SetFillStyle(fillStyle);
    }

    // Draw graph, first one with systematic errors as a box
    if (iGraph == 0) graphs[iGraph].DrawOuter("20");
    else graphs[iGraph].DrawOuter("20");
    graphs[iGraph].Draw("LPZ0");
  }

  // Update normalized graphs
  bool hasRatioPad = false;
  if (nPad >=2 && showRatioSav) {
    hasRatioPad = true;
    cd(nPad);
    if (graphsNorm.size() < graphs.size()) {
      graphsNorm.resize(0);
      for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {
        // Normalize to data
        if (dataNormPtr && hasDataSav)
          graphsNorm.push_back(VRgraphNorm(&graphs[iGraph],dataPtr));
        // Normalize to first histogram
        else {
          graphsNorm.push_back(VRgraphNorm(&graphs[iGraph],&graphs[0]));
          axesNormPtr->GetYaxis()->SetTitle("Ratio");
        }
        graphsNorm[iGraph].setColorScheme(&graphs[iGraph]);
        if (graphsNorm[iGraph].hasOuter()) {
          int fillStyle = 3354;
          if (iGraph == 1) fillStyle = 3345;
          else if (iGraph >= 2) fillStyle = 3307;
          graphsNorm[iGraph].SetFillStyle(fillStyle);
        }
      }
    }
    // Draw saved histo ratio graphs
    for (unsigned int iGraph=0; iGraph<graphsHis.size(); ++iGraph) {
      // Update
      graphsHisNorm[iGraph].update(true,ratioMinSav,ratioMaxSav);
      // Skip if no points to be plotted
      if (graphsHisNorm[iGraph].GetN() <= 0) continue;
      // Draw graph
      graphsHisNorm[iGraph].DrawOuter("20");
      graphsHisNorm[iGraph].Draw("LPZ0");
    }
    // Draw MC ratio graphs
    for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {
      // Update
      graphsNorm[iGraph].update(true,ratioMinSav,ratioMaxSav);
      // Skip if no points to be plotted
      if (graphsNorm[iGraph].GetN() <= 0) continue;
      graphsNorm[iGraph].DrawOuter("20");
      graphsNorm[iGraph].Draw("LPZ0");
    }
    // Redraw upper boundary of ratio plot in case hidden by fill styles
    xUppPtr->Draw();
  }

  // Update relative uncertainty variation pads
  int nUncVarPads = nPad - hasRatioPad -1;
  if (nUncVarPads >= 1) {
    int iPad = 1;
    for (int iGraph = 0; iGraph < int(graphs.size()); ++iGraph) {
      // Only do RelUnc pads for graphs that have (>1) outer bands
      if (!graphs[iGraph].hasOuter() || graphs[iGraph].nOuter() <= 1)
        continue;
      ++iPad;
      // Sanity check: allow for max 3 separate RelUnc pads
      if (iPad > nPad || iPad >= 3) continue;
      cd(iPad);
      // Create vector if not already done
      if (graphsVar.size() <= 1) {
        graphsVar.resize(0);
        for (int iVar=1; iVar<=graphs[iGraph].nOuter(); ++iVar) {
          graphsVar.push_back(VRgraphVar(&graphs[iGraph],iVar));
          graphsVar[graphsVar.size()-1].setStyle(iVar);
        }
      }
      for (int iVar=0; iVar<int(graphsVar.size()); ++iVar) {
        graphsVar[iVar].update();
        // Skip if no points to be plotted
        if (graphsVar[iVar].GetN() <= 0) continue;
        // Draw area-filled graphs
        if (graphsVar[iVar].GetFillStyle() != 0)
          graphsVar[iVar].Draw("e3line");
        // Draw line graphs
        else {
          graphsVar[iVar].Draw("Xline");
        }
      }
    }
  }

  // Draw objects
  // 1: lines
  cd(1);
  if (linesPtr.size() >= 1) {
    for (unsigned int iLine=0; iLine < linesPtr.size(); ++iLine)
      linesPtr[iLine]->Draw();
  }

  // Draw title
  cd(1);
  if (hasTitle) titleTexPtr->Draw();

  // Draw subtitle 1 (data ref)
  cd(1);
  if (hasDataSav) {
    if (!hasSubTitle1 && dataFilePtr->getREF() != "none")
      subTitle1("Data from "+dataFilePtr->getREF());
  }
  if (hasSubTitle1) subTitle1Ptr->Draw();

  // Draw subtitle 2 (MC names and versions)
  cd(1);
  if (hasSubTitle2) subTitle2Ptr->Draw();

  // Draw legend (so it appears on top of curves)
  cd(1);
  // Allow legend position to be redetermined so it can move.
  setLegend(legendSav);
  if (hasLegend) {
    legendPtr->Draw();
    // Optionally also draw chi2 legend
    if (hasLegendChi2 && showChi2sav) legendChi2Ptr->Draw();
  }

  // Draw legend on uncertainty band plots
  if (nUncVarPads >= 1) {
    for (int iUncVarPads=1; iUncVarPads <= nUncVarPads; ++iUncVarPads) {
      cd(1+iUncVarPads);
      setUncertaintyLabels(uLabelsSav);
      if (hasLegendUncVar) {
        legendUncVarTopPtr->Draw();
        legendUncVarBotPtr->Draw();
      }
    }
  }

  // Draw left and right labels
  cd(1);
  if (hasLeftLabel || hasRightLabel) {
    double xL = 0.+gPad->GetLeftMargin();
    double xR = 1.-0.95*gPad->GetRightMargin();
    double yT = 1.0;
    double yB = 1. - gPad->GetTopMargin();
    double yC = yB + 0.07 * (yT - yB);

    // Left label
    if (hasLeftLabel) {
      double xLm = xL + 0.01 * (xR-xL);
      if (hasLeftLabelPtr) delete leftLabelPtr;
      leftLabelPtr = new TLatex(xLm, yC, leftLabelSav.c_str());
      leftLabelPtr->SetNDC(kTRUE);
      leftLabelPtr->SetTextAlign(11);
      leftLabelPtr->SetTextColor(kBlack);
      leftLabelPtr->SetTextSize(0.04);
      leftLabelPtr->SetTextFont(42);
      leftLabelPtr->Draw();
    }

    // Right label
    if (hasRightLabel) {
      double xRm = xR - 0.01 * (xR-xL);
      if (hasRightLabelPtr) delete rightLabelPtr;
      rightLabelPtr = new TLatex(xRm, yC, rightLabelSav.c_str());
      rightLabelPtr->SetNDC(kTRUE);
      rightLabelPtr->SetTextAlign(31);
      rightLabelPtr->SetTextColor(kBlack);
      rightLabelPtr->SetTextSize(0.04);
      rightLabelPtr->SetTextFont(42);
      rightLabelPtr->Draw();
    }
  }

  // Then update canvas
  cd();
  Modified();
  Update();
  gSystem->ProcessEvents();

}

//------------------------------------------------------------------------

// Save display image

void LiveDisplay::saveLiveDisplay(string baseNameIn, string fileType) {

  // File name (without file type extension);
  string baseName = baseNameIn;
  // Revert to displayName if none provided
  if (baseNameIn == "none") {
    baseName = saveFilePrefixSav + displayNameSav + saveFileSuffixSav;
  }
  // Check which format(s) to use
  if ( fileType == "pdf" || (fileType == "def" && savePDFsav) ) {
    string fileName = baseName + ".pdf";
    Print(fileName.c_str(),"pdf");
  }
  if ( fileType == "eps" || (fileType == "def" && saveEPSsav) ) {
    string fileName = baseName + ".eps";
    Print(fileName.c_str(),"eps");
  }
  if ( fileType == "root" || (fileType == "def" && saveROOTsav) ) {
    string fileName = baseName + ".root";
    Print(fileName.c_str(),"root");
  }
  if ( fileType == "tex" || (fileType == "def" && saveTEXsav) ) {
    string fileName = baseName + ".tex";
    Print(fileName.c_str(),"tex");
  }
  if ( fileType == "png" || (fileType == "def" && savePNGsav) ) {
    string fileName = baseName + ".png";
    Print(fileName.c_str(),"png");
  }
  //
  if ( fileType == "jpeg" || (fileType == "def" && saveJPEGsav) ) {
    string fileName = baseName + ".jpeg";
    Print(fileName.c_str(),"jpeg");
  }
}

//------------------------------------------------------------------------

// save main histogram as histogram data file

bool LiveDisplay::saveHistogram(string baseNameIn, string ext) {

  // Skip 2D histograms (no save function implemented yet)
  if (has2d) return false;

  // File name (without file type extension); if "none", use displayName
  string baseName = baseNameIn;
  if (baseNameIn == "none") {
    baseName = saveFilePrefixSav + displayNameSav + saveFileSuffixSav;
  }
  string fileName = baseName + "." + ext;
  FILE * pFile = fopen (fileName.c_str(), "w");
  if (pFile == NULL) {
    perror ("(LiveDisplay::saveHistogram): Error opening file");
    return false;
  }
  else {
    // Number of bins to print
    TH1D* hisPtr = getTH1DPtr();
    int nBin = hisPtr->GetNbinsX();
    // Optional Normalization factor
    double normFactor = hisPtr->GetNormFactor();
    if (normFactor == 0.0) normFactor = 1.0;
    for (int iBin=0; iBin<=nBin+1; iBin++){
      double dx, xLo, xHi;
      // Underflow bin (save displaced by one)
      if (iBin == 0) {
        dx  = hisPtr->GetBinWidth(1);
        xLo = hisPtr->GetBinLowEdge(1)-2*dx;
      }
      // Overflow bin (save displaced by one)
      else if (iBin == nBin+1) {
        dx  = hisPtr->GetBinWidth(nBin);
        xLo = hisPtr->GetBinLowEdge(nBin)+2*dx;
      }
      // Ordinary bins
      else {
        dx  = hisPtr->GetBinWidth(iBin);
        xLo = hisPtr->GetBinLowEdge(iBin);
      }
      if ( abs(xLo/dx) < 1e-12 ) xLo = 0.;
      xHi = xLo + dx;
      if ( abs(xHi/dx) < 1e-12 ) xHi = 0.;
      double xC  = xLo + (xHi - xLo)/2.;
      double y   = normFactor * hisPtr->GetBinContent(iBin);
      // Output standard 1-sigma uncertainties
      double ey  = normFactor * hisPtr->GetBinError(iBin);
      // Systematic errors ?
      double eyThHi(0.), eyThLo(0.);
      if (graphs[0].hasOuter()) {
        for (int iVar=0; iVar < graphs[0].nOuter(); ++iVar) {
          eyThHi = max(eyThHi,
            normFactor * getTH1DPtr(0,iVar)->GetBinContent(iBin) - y);
          eyThLo = max(eyThLo,
            y - normFactor * getTH1DPtr(0,iVar)->GetBinContent(iBin));
        }
      }

      // If has systematic uncertainties
      if (graphs[0].hasOuter()) {
        fprintf(pFile,
          "%14.6g %14.6g %14.6g %14.6g %14.6g %14.6g %14.6g %14.6g \n",
          xLo,xC,xHi,y,ey,ey,eyThHi,eyThLo);
      } else {
        // Default: write out stat error only
        fprintf(pFile,"%14.6g %14.6g %14.6g %14.6g %14.6g %14.6g \n",
          xLo,xC,xHi,y,ey,ey);

      }
    }

    // Write format as text comment
    if (graphs[0].hasOuter()) {
      fprintf(pFile,"#          xLo             xC            xHi              y        +dyStat        -dyStat          +dyTh          -dyTh\n");
    }
    else {
      fprintf(pFile,"#          xLo             xC            xHi              y        +dyStat        -dyStat\n");
    }
    fprintf(pFile,"# Note! The first and last bins represent underflow and overflow, respectively (in-\n");
    fprintf(pFile,"# cluding an artificial offset by one bin width from the main histogram bins).\n");

    fprintf(pFile,"# Info:\nTIT = %s\n",hisPtr->GetTitle());
    if (hasLegend && legendSav.size() >= 1) {
      fprintf(pFile,"SRC = %s\n",legendSav[0].c_str());
    }
    if (hasSubTitle2 && subTitle2Sav != "") {
      fprintf(pFile,"REF = %s\n",subTitle2Sav.c_str());
    }
    fprintf(pFile,"#\n# File printed by LiveDisplays\n");
    fprintf(pFile,"# (Originally developed as part of the VINCIA code ");
    fprintf(pFile,"# http://arxiv.org/abs/arXiv:1102.2126)\n");
    fclose(pFile);
    return true;
  }

}

//------------------------------------------------------------------------

// Print out chi2 values

void LiveDisplay::printChi2() {
  cout<<" "<<setw(18)<<displayNameSav<<"   <chi2(5%)>  = ";
  for (int iGraph=0; iGraph<int(graphsNorm.size()); ++iGraph)
    cout<<setw(10)<<graphsNorm[iGraph].getChi2()<<" ";
  for (int iGraph=0; iGraph<int(graphsHisNorm.size()); ++iGraph)
    cout<<setw(10)<<graphsHisNorm[iGraph].getChi2()<<" ";
  cout<<endl;
}

//------------------------------------------------------------------------

// Set display title

void LiveDisplay::title(string userTitle) {
  // Can only define title if histogram has been defined
  if (!hasAxes() || graphs.size() == 0) return;
  // Get current histogram title from histogram pointer
  TH1D* hisPtr = graphs[0].getTH1DPtr();
  if (userTitle != "default") titleSav = userTitle;
  titleSav = hisPtr->GetTitle();
  // Destroy pointer if title already exists (to avoid mem leak)
  if (hasTitle) delete titleTexPtr;
  // Set placement (using axesPtr to get x and y ranges of plot)
  int NBins = axesPtr->GetNbinsX();
  double xLo = axesPtr->GetBinLowEdge(1);
  double xHi = axesPtr->GetBinLowEdge(NBins+1);
  double xTit = xLo + (xHi-xLo)/2.;
  if (gPad->GetLogx()) {
    if (xLo == 0.0) xLo += 0.5 * axesPtr->GetBinLowEdge(2);
    double lLo  = log(xLo);
    double lHi  = log(xHi);
    double lTit = lLo + (lHi - lLo)/2.;
    xTit = exp(lTit);
  }
  double yMin = axesPtr->GetMinimum();
  double yMax = axesPtr->GetMaximum();
  double dyTit = 1.16*titleSizeSav;
  double yTit = yMax - dyTit*(yMax-yMin);
  if (gPad->GetLogy()) {
    double LnyC = log(yMax) - dyTit*log(yMax/yMin);
    yTit = exp(LnyC);
  }
  // Create new title tex pointer
  titleTexPtr = new TLatex(xTit,yTit,titleSav.c_str());
  hasTitle    = true;
  titleTexPtr->SetTextFont(12);
  titleTexPtr->SetTextAlign(21);
  titleTexPtr->SetTextSize(titleSizeSav);
  titleTexPtr->SetTextColor(1);
}

//------------------------------------------------------------------------

// Set subtitle 1 (eg typically Data reference title)

void LiveDisplay::subTitle1(string subTitle1In) {
  if (!hasAxes() || graphs.size() == 0) return;
  // Destroy pointer if subTitle2 already exists (to avoid mem leak)
  if (hasSubTitle1) delete subTitle1Ptr;
  if (subTitle1In == "none") return;
  // Set placement (using axesPtr to get x and y ranges of plot)
  int NBins = axesPtr->GetNbinsX();
  double xLo = axesPtr->GetBinLowEdge(1);
  double xHi = axesPtr->GetBinLowEdge(NBins+1);
  double xTit = xLo + (xHi-xLo)/2.;
  if (gPad->GetLogx()) {
    if (xLo == 0.0) xLo += 0.5 * axesPtr->GetBinLowEdge(2);
    double lLo  = log(xLo);
    double lHi  = log(xHi);
    double lTit = lLo + (lHi - lLo)/2.;
    xTit = exp(lTit);
  }
  double yMin = axesPtr->GetMinimum();
  double yMax = axesPtr->GetMaximum();
  double dyTit = subTitle1ySav;
  double yTit = yMin + dyTit*(yMax-yMin);
  if (gPad->GetLogy()) {
    double LnyC = log(yMin) + dyTit*log(yMax/yMin);
    yTit = exp(LnyC);
  }
  subTitle1Ptr = new TLatex(xTit,yTit,subTitle1In.c_str());
  subTitle1Ptr->SetTextFont(42);
  subTitle1Ptr->SetTextAlign(21);
  subTitle1Ptr->SetTextColor(12);
  subTitle1Ptr->SetTextSize(subTitle1sizeSav);
  hasSubTitle1 = true;
}

//------------------------------------------------------------------------

// Set subtitle 2 (eg typically MC names and versions)

void LiveDisplay::subTitle2(string subTitle2In) {
  if (!hasAxes() || graphs.size() == 0) return;
  // Destroy pointer if subTitle2 already exists (to avoid mem leak)
  if (hasSubTitle2) delete subTitle2Ptr;
  // Set placement (using axesPtr to get x and y ranges of plot)
  int NBins = axesPtr->GetNbinsX();
  double xLo = axesPtr->GetBinLowEdge(1);
  double xHi = axesPtr->GetBinLowEdge(NBins+1);
  double xTit = xLo + (xHi-xLo)/2.;
  if (gPad->GetLogx()) {
    if (xLo == 0.0) xLo += 0.5 * axesPtr->GetBinLowEdge(2);
    double lLo  = log(xLo);
    double lHi  = log(xHi);
    double lTit = lLo + (lHi - lLo)/2.;
    xTit = exp(lTit);
  }
  double yMin = axesPtr->GetMinimum();
  double yMax = axesPtr->GetMaximum();
  double dyTit = subTitle2ySav;
  double yTit = yMin + dyTit*(yMax-yMin);
  if (gPad->GetLogy()) {
    double LnyC = log(yMin) + dyTit*log(yMax/yMin);
    yTit = exp(LnyC);
  }
  subTitle2Ptr = new TLatex(xTit,yTit,subTitle2In.c_str());
  subTitle2Ptr->SetTextFont(42);
  subTitle2Ptr->SetTextAlign(23);
  subTitle2Ptr->SetTextColor(12);
  subTitle2Ptr->SetTextSize(subTitle2sizeSav);
  hasSubTitle2 = true;
  subTitle2Sav = subTitle2In;
}

//------------------------------------------------------------------------

// Set Legend (align: -1: left, 0: auto, 1:right)

void LiveDisplay::setLegend(vector<string> legendIn, int alignIn) {

  // Save legend (but do not create yet if plot not yet known)
  legendSav  = legendIn;

  // Sanity check
  if (!hasAxes() || graphs.size() == 0 || legendIn.size() == 0) return;

  // Determine left/right placement
  alignSav = alignIn;
  if (alignIn == -1 || alignIn == 1) forceAlign = true;

  // Decide automatic placement based on where histogram is largest
  if (alignIn == 0) {
    double sumLeft  = 0;
    double sumRight = 0;
    // Find largest bin of data histogram (or first MC one if no data)
    int axisBins   = axesPtr->GetNbinsX();
    double axisMin = axesPtr->GetBinLowEdge(1);
    double axisMax = axesPtr->GetBinLowEdge(axisBins+1);
    double axisCtr = axisMin + 0.5*(axisMax - axisMin);
    if (gPad->GetLogx()) {
      if (axisMin == 0.0) axisMin = 0.5*axesPtr->GetBinLowEdge(2);
      double lAxisMin = log(axisMin);
      double lAxisMax = log(axisMax);
      double lAxisCtr = lAxisMin + 0.5*(lAxisMax - lAxisMin);
      axisCtr = exp(lAxisCtr);
    }
    double xMax = axisMin - 1;
    double yMax = -1e9;
    if (hasDataSav && dataPtr->GetN() >= 1) {
      int nPoints = dataPtr->GetN();
      double xTmp,yTmp;
      for (int iPoint = 0; iPoint < nPoints; iPoint++) {
        dataPtr->GetPoint(iPoint,xTmp,yTmp);
        // Do not count a point exactly in center
        if (xTmp == axisCtr) continue;
        if (yTmp > yMax) {
          yMax = yTmp;
          xMax = xTmp;
        }
      }
      // Let data count for 2 graphs:
      if (xMax > axisCtr) sumLeft += 2;
      else if (xMax >= axisMin) sumRight += 2;
    }
    if (graphs.size() >= 1) {
      for (int iGraph=0; iGraph < int(graphs.size()); ++iGraph) {
        int nPoints = graphs[iGraph].GetN();
        double xTmp,yTmp;
        // Initialize xMax to sit below axis
        xMax = axisMin - 1;
        for (int iPoint = 0; iPoint < nPoints; iPoint++) {
          graphs[0].GetPoint(iPoint,xTmp,yTmp);
          // Do not count a point exactly in center
          if (xTmp == axisCtr) continue;
          if (yTmp > yMax) {
            yMax = yTmp;
            xMax = xTmp;
          }
        }
        double w = 1.0;
        // First graph has weight 1.5 to split degeneracies
        if (iGraph == 0) w = 1.5;
        if (xMax > axisCtr) sumLeft += w;
        else if (xMax >= axisMin) sumRight += w;
      }
    }
    // Set legend position Left or Right (default right if all == 0)
    if ( sumLeft > sumRight ) alignSav = -1;
    else alignSav = 1;
  }

  // Set legend placement: right or left
  double xLeg = 0.48;
  if (!hasDataSav || !dataPtr) xLeg = 0.54;
  if (alignSav == -1) xLeg = 0.18;

  // Call setLegend with default settings
  setLegend(legendIn, xLeg);

}

//------------------------------------------------------------------------

// Set Legend with more explicit specification of position and size

void LiveDisplay::setLegend(vector<string> legendIn, double xLeg,
  double xWid, double dy, double yLegTop) {

  // Save legend for replotting
  legendSav  = legendIn;
  xLegSav    = xLeg;
  // Allow wider legend if not showing chi2 values
  if (!showChi2sav &&  nPad >= 2 ) xWid += 0.13;
  xWidSav    = xWid;
  dySav      = 2*dy;
  yLegTopSav = yLegTop;

  if (!hasAxes() || graphs.size() == 0 || legendIn.size() == 0) return;
  // nGraphs
  int nGraphsLeg = min(graphs.size(),legendIn.size());
  // Destroy pointer if already exists (to avoid mem leak)
  if (hasLegend) delete legendPtr;
  double xLegEnd = xLeg + xWid;
  int nDataLeg = hasDataSav;
  if (hasDataSav) nDataLeg = max(nDataLeg, (int)dataPtr->auxGraphs.size());
  double yLegBot = yLegTop - dy*(nGraphsLeg+nDataLeg+graphsHis.size());
  // Use two columns if lots of entries (but then disable chi2 display)
  bool useTwoColumn = false;
  if (nDataLeg > 5) {
    if (xLeg < 0.5) xLegEnd += 0.5*xWid;
    else xLeg -= 0.5 * xWid;
    useTwoColumn = true;
    showChi2sav = false;
    yLegBot = yLegTop - dy*(nGraphsLeg+nDataLeg+graphsHis.size()+1)/2;
  }
  legendPtr = new TLegend(xLeg,yLegBot,xLegEnd,yLegTop,NULL,"rNDC");
  if (useTwoColumn) legendPtr->SetNColumns(2);
  hasLegend = true;
  legendPtr->SetBorderSize(0);
  legendPtr->SetFillStyle(0);
  legendPtr->SetTextSize(0.75*dy);
  // Add data legend(s)
  if (hasDataSav) {
    if (dataPtr->auxGraphs.size() == 0) {
      string datLeg = dataFilePtr->getSRC();
      if (datLeg == "" || datLeg == "none") datLeg = "Data";
      TLegendEntry *entry=legendPtr->AddEntry("dat",datLeg.c_str(),"p");
      entry->SetMarkerStyle(dataPtr->GetMarkerStyle());
      entry->SetMarkerSize(dataPtr->GetMarkerSize());
      entry->SetMarkerColor(dataPtr->GetMarkerColor());
      entry->SetTextAlign(12);
      entry->SetTextFont(42);
      entry->SetTextColor(1);
    } else {
      map<string, VRgraph>::iterator it;
      for (it=dataPtr->auxGraphs.begin(); it!=dataPtr->auxGraphs.end(); ++it){
        string label = it->first;
        if (label == "none") label="";
        TLegendEntry *entry=legendPtr->AddEntry(label.c_str(),
          label.c_str(),"p");
        entry->SetMarkerStyle(it->second.GetMarkerStyle());
        entry->SetMarkerSize(it->second.GetMarkerSize());
        entry->SetMarkerColor(it->second.GetMarkerColor());
        entry->SetTextAlign(12);
        entry->SetTextFont(42);
        entry->SetTextColor(1);
      }
    }
  }

  // Add histogram legends
  int nLeg = 0;
  for (int iLeg = 0; iLeg < min(int(graphs.size()), int(legendIn.size()));
       ++iLeg) {
    ++nLeg;
    TLegendEntry *entry=legendPtr->AddEntry("mc",legendIn[iLeg].c_str(),"lP");
    entry->SetLineColor(graphs[iLeg].GetLineColor());
    entry->SetLineStyle(graphs[iLeg].GetLineStyle());
    entry->SetLineWidth(graphs[iLeg].GetLineWidth());
    entry->SetMarkerStyle(graphs[iLeg].GetMarkerStyle());
    entry->SetMarkerSize(graphs[iLeg].GetMarkerSize());
    entry->SetMarkerColor(graphs[iLeg].GetMarkerColor());
    entry->SetTextAlign(12);
    entry->SetTextFont(42);
    entry->SetTextColor(1);
  }
  // Add legends from saved histograms
  int nLegTot = nLeg;
  for (int iLeg = 0; iLeg < int(graphsHis.size()); ++iLeg) {
    // Safety catch
    if (iLeg >= int(hisFilePtrs.size())) continue;
    string hisLeg = hisFilePtrs[iLeg]->getSRC();
    if (hisLeg == "" || hisLeg == "none") hisLeg = "File";
    // Allow user to override auto label
    if (nLeg + iLeg < int(legendIn.size())) hisLeg = legendIn[nLeg+iLeg];
    nLegTot++;
    TLegendEntry *entry=legendPtr->AddEntry("his",hisLeg.c_str(),"lP");
    entry->SetMarkerStyle(graphsHis[iLeg].GetMarkerStyle());
    entry->SetMarkerSize(graphsHis[iLeg].GetMarkerSize());
    entry->SetMarkerColor(graphsHis[iLeg].GetMarkerColor());
    entry->SetLineColor(graphsHis[iLeg].GetLineColor());
    entry->SetLineStyle(graphsHis[iLeg].GetLineStyle());
    entry->SetLineWidth(graphsHis[iLeg].GetLineWidth());
    entry->SetTextAlign(12);
    entry->SetTextFont(42);
    entry->SetTextColor(1);
  }

  // Optionally add separate legend with chi2 values
  // (only if comparing to data)
  if (!hasDataSav || !dataNormPtr || graphsNorm.size() == 0) return;

  if (hasLegendChi2) delete legendChi2Ptr;
  xLeg    = xLegEnd + 0.004;
  // (Careful not to take this beyond right margin)
  xLegEnd = min(xLeg + 0.125, 1. - gPad->GetRightMargin() - 0.01);
  legendChi2Ptr = new TLegend(xLeg,yLegBot,xLegEnd,yLegTop,NULL,"rNDC");
  legendChi2Ptr->SetBorderSize(0);
  legendChi2Ptr->SetFillStyle(0);

  // Aligned with data legend, write: "chi2/Nbins"
  string chi2Leg = "#chi_{5%}^{2}/N_{bins}";
  TLegendEntry *entry=legendChi2Ptr->AddEntry("chi2",chi2Leg.c_str(),"");
  entry->SetTextAlign(22);
  entry->SetTextFont(42);
  entry->SetTextColor(1);

  // Insert blanks if there are several types of data points
  if (dataPtr->auxGraphs.size() > 0) {
    string blankLeg = "blank";
    for (int i=1; i<(int)dataPtr->auxGraphs.size(); ++i) {
      blankLeg += "_";
      legendChi2Ptr->AddEntry(blankLeg.c_str()," ","");
    }
  }

  // MC graphs
  int nChi2legTot = 0;
  int nChi2valid  = 0;
  for (int iLeg = 0; iLeg < int(graphsNorm.size()+graphsHis.size()); ++iLeg) {
    double chi2, chi2err;
    if (iLeg < int(graphsNorm.size())) {
      chi2 = graphsNorm[iLeg].getChi2();
      chi2err = graphsNorm[iLeg].getChi2err();
    } else {
      chi2 = graphsHisNorm[iLeg-graphsNorm.size()].getChi2();
      chi2err = graphsHisNorm[iLeg-graphsNorm.size()].getChi2err();
    }
    chi2Leg = "";
    if (chi2 >= 0.) {
      std::ostringstream s;
      s << setprecision(1);
      s << fixed << chi2 <<" #pm"<<chi2err;
      chi2Leg = s.str();
      // If chi2 > 10, just write >10
      if (chi2 >= 9.95) chi2Leg = "> 10";
      // Count up number of valid chi2 values
      nChi2valid++;
    }
    entry=legendChi2Ptr->AddEntry("chi2",chi2Leg.c_str(),"p");
    nChi2legTot++;
    entry->SetTextAlign(22);
    entry->SetTextFont(42);
    entry->SetTextColor(1);
    entry->SetTextSize(0.035);
    entry->SetMarkerStyle(21);
    entry->SetMarkerSize(0.75);
    if (chi2 < 0.) entry->SetMarkerColor(kWhite);
    else if (chi2 < 1.05) entry->SetMarkerColor(kGreen);
    else if (chi2 < 2.05) entry->SetMarkerColor(kYellow);
    else if (chi2 < 4.05) entry->SetMarkerColor(kOrange);
    else if (chi2 < 9.05) entry->SetMarkerColor(kOrange+1);
    else entry->SetMarkerColor(kRed);
  }
  // Only print chi2 values if they are aligned with main legend.
  if (nChi2legTot == nLegTot && nChi2valid >= 1) hasLegendChi2 = true;
}

//------------------------------------------------------------------------

// Set Legend for relative uncertainty variation pad

void LiveDisplay::setUncertaintyLabels(vector<string> labelsIn) {

  if (!hasAxes() || graphs.size() == 0 || labelsIn.size() == 0
    || !showRelUncSav) return;
  // Destroy pointer if already exists (to avoid mem leak)
  if (hasLegendUncVar) {
    delete legendUncVarTopPtr;
    delete legendUncVarBotPtr;
  }
  // Get axis range from axesPtr
  int NBins   = axesPtr->GetNbinsX();
  double xLo  = axesPtr->GetBinLowEdge(1);
  if (logXsav && xLo == 0.0)
    xLo += 0.5 * axesPtr->GetBinLowEdge(2);
  double xHi  = axesPtr->GetBinLowEdge(NBins+1);
  double yLoTop = 1.01;
  double yHiTop = 1.5;
  double yLoBot = -yHiTop;
  double yHiBot = -yLoTop;
  legendUncVarTopPtr = new TLegend(xLo,yLoTop,xHi,yHiTop,NULL,"r");
  legendUncVarBotPtr = new TLegend(xLo,yLoBot,xHi,yHiBot,NULL,"r");
  hasLegendUncVar = true;
  legendUncVarTopPtr->SetBorderSize(0);
  legendUncVarTopPtr->SetFillStyle(0);
  legendUncVarTopPtr->SetNColumns(7);
  legendUncVarBotPtr->SetNColumns(7);
  legendUncVarBotPtr->SetBorderSize(0);
  legendUncVarBotPtr->SetFillStyle(0);
  // Add uncertainty variation legends
  for (int iLeg = 0; iLeg < int(labelsIn.size()); ++iLeg) {
    TLegend* legPtr = (iLeg < 5) ? legendUncVarTopPtr : legendUncVarBotPtr;
    if (iLeg >= int(graphsVar.size())) continue;
    string opt = "l";
    if (graphsVar[iLeg].GetFillStyle() != 0.) opt = "lf";
    TLegendEntry *entry=legPtr->AddEntry("RU",labelsIn[iLeg].c_str(),
      opt.c_str());
    entry->SetLineColor(graphsVar[iLeg].GetLineColor());
    entry->SetLineStyle(graphsVar[iLeg].GetLineStyle());
    entry->SetLineWidth(graphsVar[iLeg].GetLineWidth());
    entry->SetMarkerStyle(graphsVar[iLeg].GetMarkerStyle());
    entry->SetMarkerSize(graphsVar[iLeg].GetMarkerSize());
    entry->SetMarkerColor(graphsVar[iLeg].GetMarkerColor());
    entry->SetFillStyle(graphsVar[iLeg].GetFillStyle());
    entry->SetFillColor(graphsVar[iLeg].GetFillColor());
    entry->SetTextAlign(12);
    entry->SetTextFont(42);
    entry->SetTextColor(1);
  }
}

//==========================================================================

// Simple utility to convert a Pythia Hist to a ROOT TH1 object.
// Note that statistical uncertainties are not preserved.

void Hist2TH1(TH1* th1Ptr, Pythia8::Hist* histPtr) {
  for (int iBin = 0; iBin <= th1Ptr->GetNbinsX() + 1; ++iBin) {
    double cont = histPtr->getBinContent(iBin);
    th1Ptr->SetBinContent(iBin,cont);
  }
}

} // End namespace Pythia8
