// VinciaMG5cxxInterface.cc is a part of the PYTHIA event generator.
// Copyright (C) 2019 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL v2 or later, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// Main interface methods for MadGraph 5 C++ matrix elements
// (MEs generated with the MG5 PY8Kernels Plugin)

#include "Pythia8/VinciaMG5cxxInterface.h"

namespace Pythia8 {

//============================================================================

// The mg5cxx interface

//----------------------------------------------------------------------------

// Main routine to initialise the MadGraph coupling common blocks, required
// by the MG matrix elements. Adapted from MG5 : SOURCE/MODEL/couplings.f
// Also set global information to be used in calls to MG5 matrix elements.

bool VinciaMG5cxxInterface::init() {

  // Set verbose level
  verbose = settingsPtr->mode("Vincia:verbose");

  // Verbose output
  if (verbose > normal) printOut("VinciaMG5cxxInterface::init","...");

  // Check if pointers initialized
  if (! isInitPtr) {
    printOut("VinciaMG5cxxInterface::init",
      "Cannot initialize, pointers not set.");
    return false;
  }

  // Innocent until proven guilty
  isInit = true;

  // Set colour depth (todo: via "Vincia:matchingFullColour")
  bool fullColour = true;
  colourDepth = (fullColour) ? 1 : 0;

  // Initialise Parameters_sm (only if not using dummy)
#ifdef MG5MES

  // Verbose output
  if (verbose >= 3) printOut("VinciaMG5cxxInterface::init",
    "   setting MG C++ masses, widths, couplings...");

  // Create new model instance.
  Parameters_sm* modelPtr = new Parameters_sm();

  modelPtr->setIndependentParameters(particleDataPtr,couplingsPtr,slhaPtr);
  modelPtr->setIndependentCouplings();
  if (verbose >= 3) {
    modelPtr->printIndependentParameters();
    modelPtr->printIndependentCouplings();
  }

  // In the VINCIA context, alphaS_MGME = 1/4pi (- > gS = 1; we
  // control the running separately). Thus, even the MG5 "dependent"
  // parameters only need to be set once.

  // Alternatively, we could evaluate the QCD coupling at MZ but should
  // then use a coupling definition from a Vincia parameter list rather
  // than PYTHIA's couplings.
  //    double muDefME  = 91.2;
  //    double alpS = couplingsPtr->alphaS(pow2(muDefME));

  // The following is equivalent to
  // PY8MEs::updateModelDependentCouplingsWithPY8(...)
  double alpS = 1.0 / ( 4 * M_PI );
  modelPtr->setDependentParameters(particleDataPtr, couplingsPtr, slhaPtr,
    alpS);
  //model.setDependentParameters();
  modelPtr->setDependentCouplings();

  // Construct MG5 process library
  if (verbose >= 9) {
    printOut("VinciaMG5cxxInterface::init()",
      "   attempting to construct mg5lib");
  }
  if (mg5libPtr != NULL && mg5libPtr != 0) delete mg5libPtr;
  mg5libPtr = new PY8MEs(modelPtr);
#else
  mg5libPtr = NULL;
#endif

  return true;
}

//--------------------------------------------------------------------------

// Get Matrix Element Squared
// The first nIn particles must be the incoming ones
// If MG5MES plugin not linked, pure dummy implementation instead.

#ifdef MG5MES

double VinciaMG5cxxInterface::ME2(vector<Particle> state, int nIn) {

  // Sanity checks
  if (nIn <= 0) return -1;
  else if (state.size() - nIn < 1) return -1;

  // Prepare vector of incoming ID codes
  vector<int> idIn, helOrg, col, acol;
  vector<Vec4> momenta;
  idIn.push_back(state[0].id());
  momenta.push_back(state[0].p());
  helOrg.push_back(state[0].pol());
  col.push_back(state[0].col());
  acol.push_back(state[0].acol());
  if (nIn == 2) {
    idIn.push_back(state[1].id());
    momenta.push_back(state[1].p());
    helOrg.push_back(state[1].pol());
    col.push_back(state[1].col());
    acol.push_back(state[1].acol());
  }
  // Prepare vector of outgoing ID codes
  vector<int> idOut;
  for (int i=nIn; i<(int)state.size(); ++i) {
    idOut.push_back(state[i].id());
    momenta.push_back(state[i].p());
    helOrg.push_back(state[i].pol());
    col.push_back(state[i].col());
    acol.push_back(state[i].acol());
  }
  // Currently not using the option to request specific s-channels
  set<int> sChannels;

  // Access the process
  process_specifier proc_spec = mg5libPtr->getProcessSpecifier(idIn, idOut,
    sChannels);
  process_accessor proc_handle = mg5libPtr->getProcess(proc_spec);

  // Return right away if unavailable
  if (proc_handle.second.second < 0) return -1;

  // Convert momenta and colours to MG5 format
  vector< vector<double> > momentaMG5;
  vector< int > colacolMG5;
  for (int i = 0; i<(int)momenta.size(); ++i) {
    vector<double> pNow;
    // MG5 expects energy as first entry.
    pNow.push_back(momenta[i].e());
    pNow.push_back(momenta[i].px());
    pNow.push_back(momenta[i].py());
    pNow.push_back(momenta[i].pz());
    momentaMG5.push_back(pNow);
    colacolMG5.push_back(col[i]);
    colacolMG5.push_back(acol[i]);
  }

  vector<int> i9;
  // Check if we are doing a (partial) helicity sum?
  for (int i=0; i<(int)helOrg.size(); ++i) {
    // Save indices of unpolarised partons
    if (helOrg[i] == 9) i9.push_back(i);
  }

  // Explicitly sum over any hel = 9 helicities
  vector< vector<int> > helConf;
  helConf.push_back(helOrg);
  while (i9.size() > 0) {
    int i  = i9.back();
    int id = (i < nIn) ? idIn[i] : idOut[i-nIn];
    // How many spin states?
    int nS = particleDataPtr->spinType(id);
    // Massless particles max have max 2 (physical) spin states
    if (particleDataPtr->m0(id) == 0.0) nS=min(nS,2);
    // Create nS copies of helConf, one for each spin state
    int helConfSizeNow = helConf.size();
    for (int iCopy=1; iCopy <= nS; ++iCopy) {
      // Set hel for this particle in this copy.
      // Start from -1, then 1, then 0 (if 3 states)
      int h = -1;
      if (iCopy == 2) h = 1;
      else if (iCopy == 3) h = 0;
      else if (iCopy == 4) h = -2;
      else if (iCopy == 5) h = 2;
      for (int iHelConf=0; iHelConf<helConfSizeNow; ++iHelConf) {
        vector<int> helNow = helConf[iHelConf];
        helNow[i] = h;
        // First copy: use existing
        if (iCopy == 1) helConf[iHelConf] = helNow;
        // Subsequent copies: create new
        else helConf.push_back(helNow);
      }
    }
    // Remove the particle whose helicities have been summed over
    i9.pop_back();
  }

  // VERBOSE DEBUG
  if (verbose >= 9) {
    cout<<" in = ";
    for (int i=0; i<(int)idIn.size(); ++i) cout<<idIn[i]<<" ";
    cout<<"   out = ";
    for (int i=0; i<(int)idOut.size(); ++i) cout<<idOut[i]<<" ";
    cout<<endl;
    cout<<" number of helicity configurations = "<<helConf.size()<<endl;
    for (int i=0; i<(int)helConf.size(); ++i) {
      cout<<"   helConf "<<i;
      for (int j=0; j<(int)helConf[i].size(); ++j) {
        cout<<" "<<helConf[i][j];
      }
      cout<<endl;
    }
  }

  // Set properties and return ME2 value
  PY8ME* proc_ptr = proc_handle.first;
  vec_int perms = proc_handle.second.first;
  int proc_ID = proc_handle.second.second;

  proc_ptr->setMomenta(momentaMG5);
  proc_ptr->setProcID(proc_ID);
  proc_ptr->setPermutation(perms);
  proc_ptr->setColors(colacolMG5);

  // Compute helicity sum (and save helicity components if needed later)
  double me2 = 0.0;
  me2hel.clear();
  for (int iHC=0; iHC<(int)helConf.size(); ++iHC) {

    // Note: could check here if the helConf is physical
    // (consistent with spins of particles and conserving angular momentum)

    // Verbose output: print helicity configuration in Vincia's enumeration
    //cout<<" helConfNow = "<<iHC<<" size = "<<helConf[iHC].size();
    //for (int h=0; h<(int)helConf[iHC].size(); ++h)
    //  cout<<" "<<helConf[iHC][h];
    //cout<<endl;

    proc_ptr->setHelicities(helConf[iHC]);
    double me2now = proc_ptr->sigmaKin();
    // MG may produce inf/nan for unphysical hel combinations
    if ( !std::isnan(me2now) && !std::isinf(me2now) ) {
      // Save helicity matrix element for possible later use.
      me2hel[helConf[iHC]] = me2now;
      // Add this helicity ME to me2
      me2 += me2now;
    }


  }

  me2*=double(proc_ptr->getSymmetryFactor());
  return  me2;

}

#else

// Pure dummy implementation if MG5MES plugin not linked.

double VinciaMG5cxxInterface::ME2(vector<Particle>, int) {
  return -1.;
}

#endif

//--------------------------------------------------------------------------

// Set helicities for a particle state
// (Note: takes a reference as input and operates on it)

bool VinciaMG5cxxInterface::selectHelicities(vector<Particle>& state, int nIn){

  // Get the matrix element (automatically sums over any h=9 particles)
  double me2sum = ME2(state, nIn);
  if (verbose >= 9)
    cout<<" VinciaMG5cxxInterface::selectHelicities(): "
        <<scientific<<me2sum<<endl;

  // Did we find the ME? ( ME2() returns -1 if no ME found ).
  if (me2sum <= 0.) return false;

  // Check how many helicity configurations we summed over
  int nHelConf = me2hel.size();
  if (nHelConf <= 0) return false;

  // Random number between zero and me2sum (trivial if only one helConf)
  double ranHelConf = 0.0;
  vector<int> hSelected;
  if (nHelConf >= 2) ranHelConf = rndmPtr->flat() * me2sum;
  for (map< vector<int>, double>::iterator it=me2hel.begin();
       it != me2hel.end(); ++it) {
    // Progressively subtract each me2hel and check when we cross zero
    ranHelConf -= it->second;
    if (ranHelConf <= 0.0) {
      hSelected = it->first;
      break;
    }
  }
  if (ranHelConf > 0.) return false;

  // Set helicity configuration
  for (int i=0; i<(int)state.size(); ++i) state[i].pol(hSelected[i]);

  // Verbose output
  if (verbose >= 9)
    cout<<" VinciaMG5cxxInterface::selectHelicities(): selected "<<
      makeLabel(hSelected, nIn, false)<<endl;

  // Everything OK
  return true;
}

//--------------------------------------------------------------------------

// Convert process id codes (or helicity values) to string

string VinciaMG5cxxInterface::makeLabel(vector<int>& id, int nIn,
  bool useNames) {

  string label = "{";
  for (int i=0; i<(int)id.size(); ++i) {
    string idName;
    if (useNames && id[i] != 0) idName = particleDataPtr->name(id[i]);
    else idName = num2str(id[i]);
    if (i == nIn-1) idName += " ->";
    label += idName+" ";
  }
  label += "}";
  return label;
}



} // end namespace Pythia8
